%% CONVERTER INFO.
%--------------------------------------------------------------------------
%  Intro.
%
%    This section describes how a convertion function ("convert_Function")
%    must behave. The point of the convertion function is to: tell the
%    converter GUI how to display and describe the converter and which
%    convertion options to offer the user; and to perform the actual
%    convertion, and return data if requested. The desired action is
%    toggled by calling the function using a flags in the first argument.
%
%    The converter function should throw errors when it encounters them.
%
%    Note, the converter only supports raw files with one eye-tracking
%    dataset.
%
%    Note, the conversion function may be called in parallel (not
%    yet implemented).
%
%    
%--------------------------------------------------------------------------
%  Calling convention:
%
%    varargout = convert_Function(doWhat, varargin)
%
%      where doWhat indicates the necessary action. The varagin and
%      varargout depend on the action:
%
%        # s = convert_Function('getConverterInfo')
%        # s = convert_Function('setupConverter')
%        # s = convert_Function('getFileInfo', setts)
%        #     convert_Function('convertFile', setts)
%        # s = convert_Function('getData', setts)
%
%
%--------------------------------------------------------------------------
%  getConverterInfo:
%    This function is called with doWhat = 'getConverterInfo' when the
%    FileConverter is requesting info about the converter. This mode should
%    return the name and description of the convert; and, optionally, setup
%    options for the converter.
%
%    Calling convention:
%      s = convert_Function('getConverterInfo');
%
%        where the output s is a struct containing the following fields:
%          # name:  The name of the converter.
%          # desc:  The description of the converter.
%
%        Additionally, s can contain a field 'setupSetts', which features
%        custom settings for setting up the converter. The user selection
%        options will then be returned to this function when its called
%        with the setupConverter flag. It can then use those settings to
%        e.g. set the extension of the converter. As such, a converter can
%        support different types of extensions, but once chose, it remains
%        bound to it.
%
%
%--------------------------------------------------------------------------
%  setupConverter:
%    This function is called when a certain converter is chosen, and should
%    return the final immutable settings for the converter, such as the
%    extension of the files it will convert. This extension is then used by
%    the Converter GUI to allow the users to select files. This mode can
%    also return a definition of the necessary converter settings, which
%    will be returned to this function when called in the convert mode.
%
%    Calling convention:
%      s = convert_Function('setupConverter',setts);
%
%        where the input setts is a struct containing the fields defined in
%        setupSetts when this function is called with the getConverterInfo
%        flag, see above. Each field contains, in a value subfield, the
%        options chosen by the user.
%
%        and where the output s is a struct containing the following field:
%          # ext :  The extension of the files.
%
%        Additionally, s can contain a field 'customSetts', which features
%        custom settings for the converter. When the file converter then
%        calls the converter function with convertFile flag, it adds a
%        customSetts field to the setts struct, wherein the fields from the
%        original customSetts is included. Those fields contain only the
%        value field, with the value as set by the user (or left at
%        default).
%
%
%--------------------------------------------------------------------------
%  getFileInfo:
%    This function is called with doWhat = 'getFileInfo' when the
%    FileConverter (re)loads a folder: for each file with the extension as
%    returned by a call with doWhat = 'getConverterInfo', this function is
%    called.
%
%    Calling convention:
%      s = convert_Function('getFileInfo',setts);
%
%        where the struct setts includes all fields mentioned in all
%        abovementioned modes, as well as the raw-file name in the
%        'rawFileName' field. This function should only return metadata
%        about the file.
%
%        and where the output s is a struct containing the following
%        fields, which correspond the the PhysioData convention:
%          > For Signals:
%              # s.signals.channelNames:  Cell array of channel names.
%              # s.signals.channelUnits:  Cell array of channel units.
%              # s.signals.fs          :  The sampling rate, in Hz.
%              # s.signals.L           :  The signal length, in samples.
%
%          > For Eye-Tracking Datasets:
%              # s.eyeTracking.nL         : Num of valid left (L) samples.
%              # s.eyeTracking.nR         : Num of valid right (R) samples.
%              # s.eyeTracking.nLR        : Num of valid L or R samples
%              # s.eyeTracking.elapsedTime: The ET of the recording.
%              # s.eyeTracking.diamFs     : Approx. fs of the diam. rec.
%
%          > For Labels:
%              # s.labels.count: The number of labels in the file.
%
%
%--------------------------------------------------------------------------
%  convertFile:
%    When called with doWhat = 'convertFile', the function converts a
%    supported raw file into a physiOData file.
%
%    Note, the user is not given the option to enable/disable eye-tracking
%    data for extraction. Instead, the eye-tracking converter function must
%    decide itself which data to convert.
%
%    Calling convention:
%      convert_Function('convertFile',setts);
%
%       where setts is a struct containing all fields mentioned in all
%       abovementioned modes, as well as the following fields:
%         # doAppend            :  Flag indicating if the saving should
%                                  occur using the append flag. The File
%                                  Converter checks if the user requested
%                                  appending AND if the file already
%                                  exists, if so, the doAppend flag is set
%                                  to true. It is otherwise false.
%         # signalsToExtract    :  The one-based indxs of sigs to extract.
%         # newFileName         :  The full filename of the to-be-produced
%                                  PhysioData file.
%         # rawFileName         :  The full filename of the to-be-converted
%                                  raw file.
%         # customSetts         :  The custom settings, see above.
%         # includeRawCopy      :  Flag indicating if a copy of the raw
%                                  data should be included in the UserData
%                                  field. Currently implemented in the
%                                  Eye-Tracking data.
%
%    Note: The function should not return anything on success, and should
%    throw an error on fail. The message of the error will be exposed to
%    the user.
%
%
%--------------------------------------------------------------------------
%  getData:
%    When called with doWhat = 'getData', the function should return the
%    data from a raw file. The inputs are the same as with the
%    'convertFile' flag, but excluding the doAppend field.
%
%    Unlike calling the function with the 'convertFile' file, calling it
%    with 'getData' does not create a file, but instead returns the file
%    data as a struct. See the PhysioData file specification in the User
%    Guide.
%
%    Calling convention:
%      s = convert_Function('getData',setts);
%
%
%--------------------------------------------------------------------------
%  The File Converter can present two types of options to the user, setup
%  options and converter options, these options must be defined via the
%  getConverterInfo and setupConver modes of this function, see above.
%  
%  A struct defining a set of options/settings must conform to the format
%  below:
%        struct must contain the fields:
%           # type   : the type of setting:
%                        'choice' -> radio buttons. 'check'  -> checkboxes.
%                        'text'   -> free-write input box.
%           # value  : the def value of the setting (choice/check number or
%                      text string).
%           # label  : the label of the setting.
%           # desc   : the description the setting.
%           # choices: The choices, only with types choice and check.
%
%
%--------------------------------------------------------------------------
% Elio Sjak-Shie, July 2019
%--------------------------------------------------------------------------
















