classdef EyeTrackingDataModel < handle
    % EyeTrackingData - Class for managing eye tracking data.
    %
    %   The class presents a data model for generating eye-tracking data
    %   compatible with the PhysioData Toolbox.
    %
    %----------------------------------------------------------------------
    %   Part of the PhysioData Toolbox.
    %    Elio Sjak-Shie, Faculty of Social Sciences,
    %     Leiden University, 2017.
    %----------------------------------------------------------------------
    
    
    %% PhysioData Properties:

    properties
        
        % filename - the name of the source file.
        filename     = '';
        
        % name - name of the ET dataset.
        name         = 'NAME';
                
        % diameterUnit - The unit of the diameter.
        diameterUnit = 'mm';
        
        % raw_t_ms_max - double for holding the maximum time in the raw
        % dataset.
        raw_t_ms_max = [];
        
        % gaze - n x 1 struct containing the n snapshots and their data.
        % The data consists of the snapshots: name, image, width, and
        % height. Additionally, the Gaze coordinates, fixation coordinates,
        % and, RoI hits, and metadata are saved. The RoI-struct is a k x 1
        % struct--for k RoI regions--containing the RoI's name in 'name',
        % and the start and end times of the hits, in milliseconds, in the
        % 'hit' field. The  h hits are saved in h x 2 matrix with the start
        % and end times as columns. The gazePoints and fixationPoints
        % fields are struct with times, in ms, and points, in XY
        % cooridnates.
        gazeData = struct('snapshotName',''...
            ,'snapshotImage',[]...
            ,'snapshotWidth',[]...
            ,'snapshotHeight',[]...
            ,'gazePoints',struct('t_ms',[],'XY',[])...
            ,'fixationPoints',struct('t_ms',[],'XY',[])...
            ,'RoI',struct('name',{},'hit',{})...
            ,'UserData',[]);
        
        % diameter - a scalar struct containing the time vector, in ms, and
        % the raw measured pupil diameters of the left and/or right eye.
        % The diameters should be in millimeters.
        diameter     = struct('t_ms',[],'L',[],'R',[]);
        
        % eventSections - n x 1 struct that holds n scored events. These
        % events are scored visually and are independent of the snapshots.
        % The struct is similar to the RoI struct in gazeData.
        eventSections = struct('name',{},'hit',{});
        
        % labels - labels.
        labels        =  struct('name',{''},'t_ms',{[]});
        
        % metadata -  blank struct for containing extra metadata.
        UserData      = [];
        
    end
    
    
    
    %% Constructor:
    
    methods
        
        %==================================================================
        function obj = EyeTrackingDataModel()
            % Constructor
            
        end
        
    end
    
    
    %% Get/Set and Error Checking Functions:
    
    methods
        
        
        %==================================================================
        function set.filename(obj,valIn)
            
            obj.dataCheck('filename',valIn);
            obj.filename = valIn;
            
        end
        
        
        %==================================================================
        function set.name(obj,valIn)
            
            obj.dataCheck('name',valIn);
            obj.name = valIn;
            
        end

        
        
        %==================================================================
        function set.diameterUnit(obj,valIn)
            
            obj.dataCheck('diameterUnit',valIn);
            obj.diameterUnit = valIn;
            
        end
        
        
        %==================================================================
        function set.raw_t_ms_max(obj,valIn)
            
            obj.dataCheck('raw_t_ms_max',valIn);
            obj.raw_t_ms_max = valIn;
            
        end
        
        
        %==================================================================
        function set.gazeData(obj,valIn)
            
            obj.dataCheck('gazeData',valIn);
            obj.gazeData = valIn;
            
        end
        
        
        %==================================================================
        function set.diameter(obj,valIn)
            
            obj.dataCheck('diameter',valIn);
            obj.diameter = valIn;
            
        end
        
        
        %==================================================================
        function set.eventSections(obj,valIn)
            
            obj.dataCheck('eventSections',valIn);
            obj.eventSections = valIn;
            
        end
        
        
        %==================================================================
        function set.labels(obj,valIn)
            
            obj.dataCheck('labels',valIn);
            obj.labels = valIn;
            
        end
        
        
        %==================================================================
        function dataCheck(~,propName,valIn)
            
            switch propName
                
                case 'name'
                    assert(~isempty(valIn) && ischar(valIn)...
                        ,['The name prop must be char,'...
                        ' and cannot be empty.']);
                        
                case 'filename'
                    assert(ischar(valIn)...
                    ,['The filename prop must'...
                    ' be empty or chars.']);
                
                case 'diameterUnit'
                    assert(strcmp(valIn,'mm') || strcmp(valIn,'px')...
                        ,['The PhysioData Toolbox currently only'...
                        ' supports'...
                        ' ''mm'' or ''px'' as the pupil diameter.']);
                    
                case 'raw_t_ms_max'
                    assert(isempty(valIn) ...
                        || (isnumeric(valIn) && isscalar(valIn))...
                        ,'raw_t_ms_max must be numeric (possibly []).');
                    
                case 'gazeData'
                    
                    % Allow the gazeData to be empty:
                    if isempty(valIn)
                        return
                    end
                    
                    for curGazeF = valIn(:)'
                        
                        % The name Width and Height must be present:
                       assert(...
                           all(isfield(curGazeF...
                           ,{'snapshotName'...
                           'snapshotWidth'...
                           'snapshotHeight'...
                           }...
                           )));
                       
                       % Check the optional fields:
                       if isfield(curGazeF,'snapshotImage')
                           assert(...
                               isempty(curGazeF.snapshotImage)...
                               || isa(curGazeF.snapshotImage,'uint8')...
                               ,['The snapshotImage field must be'...
                               ' empty or a uint8.'])
                       end
                          
                       TestPointsStruct = @(s) ...
                            isempty(s)...
                               || ...
                               (...
                               ...
                               isstruct(s)...
                               && all(isfield(s,{'t_ms' 'XY'}))...
                               ...
                               && isequal(s.t_ms,s.t_ms,[])...
                               ...
                               || ...
                               (...
                               isnumeric(s.t_ms) && size(s.t_ms,2) == 1 ...
                               && isnumeric(s.XY) && size(s.XY,2) == 2 ...
                               && size(s.t_ms,1) == size(s.XY,1) ...
                               )...
                               ...
                               );
                           if isfield(curGazeF,'gazePoints')
                               assert(...
                                   TestPointsStruct(...
                                   curGazeF.gazePoints)...
                                   ...
                                   ,['The gazePoints field is not'...
                                   ' correct, read User Guide.']);
                           end
                           if isfield(curGazeF,'fixationPoints')
                               assert(...
                                   TestPointsStruct(...
                                   curGazeF.fixationPoints)...
                                   ...
                                   ,['The fixationPoints field is not'...
                                   ' correct, read User Guide.']);
                           end
                           
                           if isfield(curGazeF,'RoI')
                               assert(...
                                   isempty(curGazeF.RoI)...
                                   || (...
                                   isstruct(curGazeF.RoI)...
                                   && all(isfield(curGazeF.RoI...
                                   ,{'name' 'hit'}))...
                                   )...
                                   ...
                                   ,['The RoI field is not'...
                                   ' correct, read User Guide.']);
                               assert(all(cellfun(...
                                   @(cc) ischar(cc)...
                                   ,{curGazeF.RoI.name}))...
                                   ,['The name field in the RoI field'...
                                   ' must contain a char,'...
                                   ' see User Guide.']);
                               assert(all(cellfun(...
                                   @(cc) isnumeric(cc)...
                                   ,{curGazeF.RoI.hit}))...
                                   ,['The hit field in the RoI field'...
                                   ' must be numeric,'...
                                   ' see User Guide.']);
                               assert(all(cellfun(...
                                   @(cc) size(cc,2)==2 ...
                                   ,{curGazeF.RoI.hit}))...
                                   ,['The hit field in the RoI field'...
                                   ' must be a m x 2 matrix,'...
                                   ' see User Guide.']);
                           end                        
                    end        
                    
                case 'diameter'
                    
                    % The diameter struct must contain a t_ms field that is
                    % a numeric strickly monotonically increasing vector,
                    % and at least a L or R field:
                    assert(...
                        isfield(valIn,'t_ms') ...
                        && isnumeric(valIn.t_ms) ...
                        && all(diff(valIn.t_ms)>0)...
                        && size(valIn.t_ms,2) == 1 ...
                        ,['diameter must have a t_ms field that is'...
                        ' a numeric and strickly'...
                        ' increasing single-column vector.']);
                    assert(...
                        any(isfield(valIn,{'L' 'R'})) ...
                        ,'diameter must have an L or R field.');
                    tSize = length(valIn.t_ms);
                    
                    % If present and non-empty, the L or R fields must be
                    % the same size as t_ms:
                    checkPupilData = @(f) ...
                        assert(...
                        ~isfield(valIn,f) ...
                        || ...
                        isequal(valIn.(f),[]) ...
                        || ...
                        (size(valIn.(f),2) == 1 ...
                        && size(valIn.(f),1) == tSize ...
                        )...
                        ...
                        ,['The ' f ' field, if present and non-empty,'...
                        ' must have a single column'...
                        ' and as many rows as the t_ms field.']);
                    
                    checkPupilData('L');
                    checkPupilData('R');
                    
                case 'eventSections'
                    
                    for curEvt = valIn(:)'
                        
                        assert(...
                            (...
                            isstruct(curEvt)...
                            && all(isfield(curEvt,{'name' 'hit' 'desc'}))...
                            && ischar(curEvt.name)...
                            && ischar(curEvt.desc)...
                            && isnumeric(curEvt.hit)...
                            && size(curEvt.hit,2) == 2 ...
                            && all(curEvt.hit(:,2) > curEvt.hit(:,1))...
                            )...
                            ...
                            ,['The eventSections field is not correct'...
                            ', read User Guide.']);
                        
                    end
                    

                    
                    
                case 'labels'
                    
                    assert(...
                        isempty(valIn)...
                        || (...
                        isstruct(valIn)...
                        && all(isfield(valIn,{'name' 't_ms'}))...
                        ...
                        && ...
                        (...
                        (isempty(valIn.name) && isempty(valIn.t_ms))...
                        ||...
                        (...
                        iscell(valIn.name)...
                        && isnumeric(valIn.t_ms)...
                        && size(valIn.name,2) == 1 ...
                        && size(valIn.t_ms,2) == 1 ...
                        && size(valIn.name,1) == size(valIn.t_ms,1) ...
                        )...
                        )...
                        )...
                        ,['The label field is not correct'...
                        ', read User Guide.']);
                    
                    
                otherwise
                    error('Error!');
            end
        end
        
        
        
        %==================================================================
        function timeShiftData(obj, offsetSec)
            
            % Shift the signal if an offset is given:
            if nargin>1

                % Offset:
                offset_ms = offsetSec*1000;
                
                % Process diameter data:
                obj.diameter.t_ms = obj.diameter.t_ms+offset_ms;
                
                % Process gazeData:
                for ii = 1: length(obj.gazeData)
                    obj.gazeData(ii).gazePoints.t_ms ...
                        = obj.gazeData(ii).gazePoints.t_ms+offset_ms;
                    obj.gazeData(ii).fixationPoints.t_ms ...
                        = obj.gazeData(ii).fixationPoints.t_ms+offset_ms;
                    for kk = 1: length(obj.gazeData(ii).RoI)
                        obj.gazeData(ii).RoI(kk).hit...
                            = obj.gazeData(ii).RoI(kk).hit + offset_ms;
                    end
                end
                clear ii kk
                
                % Process eventSections:
                for ii = 1:length(obj.eventSections)
                    obj.eventSections(ii).hit ...
                        =  obj.eventSections(ii).hit + offset_ms;
                end
                clear ii
                
                % Process labels:
                obj.labels.t_ms = obj.labels.t_ms+offset_ms;
                
                % Process raw_t_ms_max:
                obj.raw_t_ms_max = obj.raw_t_ms_max+offset_ms;
                
                % Add description:
                fprintf(['Original signal shifted by'...
                    ' %.4f s.\n'],offsetSec);
                
            end
            
            
        end
        
        
        %==================================================================
        function s = returnStruct(objArray)
            % Returns a data struct containing the eye-tracking raw data.
            %
            %  s = returnStruct(obj) return the raw data struct with the
            %  original, zeroed time vector.
            %
            %  s = returnStruct(obj, offsetSec) returns the raw data struct
            %  with the times offset by offsetSec seconds. Note that the
            %  data is not sectioned, only offset, as such, negative times
            %  may arise.
            %
            
            % Allow array call:
            for objIndx = numel(objArray) : -1 : 1
                
                obj = objArray(objIndx);
                
                % Collect the fields to export:
                for f = {'filename' 'name' 'diameter' 'gazeData'...
                        'eventSections' 'labels' 'diameterUnit' ...
                        'raw_t_ms_max' 'UserData'}
                    s(objIndx).(f{1}) = obj.(f{1});
                end
            end
            
            
        end
        
        
    end
    
end