classdef PhysioDataFileModel < handle
    % PhysioDataFileModel Data-Model class for managing PhysioData files.
    %
    %  See included example files and the User Guide for details.
    %
    %----------------------------------------------------------------------
    %   Part of the PhysioData Toolbox.
    %    Elio Sjak-Shie, Faculty of Social and Behavioural Sciences,
    %     Leiden University, 2017.
    %----------------------------------------------------------------------
    
    
    %% Properties:
    
    properties
        data           = [];
        epochs         = [];
        physioDataInfo = [];
    end
    
    %NOTE this class needs to be rewritten.
    
    %TODO (HIGH) Include UserData field.
    
    
    %% Methods:
    
    methods
        
                    
        %==================================================================
        function obj = PhysioDataFileModel()
            % Constructor.
            
        end
        
        
        %==================================================================
        function set.data(obj,valIn)
            % Checks the data property.
            
            if isempty(valIn)
                obj.data = valIn;
                return
            end
            
            assert(isstruct(valIn) && isscalar(valIn)...
                ,'The data prop must be a scalar stuct.');
%             assert(any(isfield(valIn,{'signals' ...
%                 'eyeTracking'}))...
%                 ,['The signals field, if present, must contain the'...
%                 ' a ''signals'' field and/or a ''eyeTracking'' field.']);
            
            % Check signals:
            if isfield(valIn,'signals')
                assert(all(isfield(valIn.signals,{'fs' 'channels' ...
                    'channelUnits' 'channelNames' 'channelDescription'}))...
                    ,['The signals field, if present, must contain the'...
                    ' following fields: ''fs'' ''channels''' ...
                    '''channelUnits'' ''channelNames'''...
                    'and ''channelDescription''.']);
                
                % Check channels:
                nChans   = size(valIn.signals.channels,2);
                assert(iscell(valIn.signals.channels) ...
                    && size(valIn.signals.channels,1)==1 ...
                    ,['The channels field must be a cell'...
                    ' array or scalar with a single row.'])
                nSamples = size(valIn.signals.channels{1},1);
                assert(all(cellfun( @(d) ...
                    size(d, 2) == 1 && size(d, 1) == nSamples ...
                    ,valIn.signals.channels))...
                    ,['All channels must have a single column'...
                    ', and the same number of samples.'])
                
                % Check fs:
                assert(isnumeric(valIn.signals.fs) ...
                    && valIn.signals.fs > 0 && valIn.signals.fs < 100000 ...
                    ,['The fs must be numeric and greater than 0'...
                    ' , but less than 100000.'])
                
                % Check other fields:
                assert(...
                    iscellstr(valIn.signals.channelUnits) ...
                    && iscellstr(valIn.signals.channelNames) ...
                    && iscellstr(valIn.signals.channelDescription) ...
                    ,['The channelUnits channelNames and'...
                    ' channelDescription fields must be cellstrings.']);
                    
                assert(...
                    isequal(...
                    size(valIn.signals.channelUnits) ...
                    ,size(valIn.signals.channelNames)...
                    ,size(valIn.signals.channelDescription)...
                    ,[1 nChans]) ...
                    ,['The channelUnits channelNames and'...
                    ' channelDescription fields must have one column'...
                    ' per channel.']);
                
            end
            
            if isfield(valIn,'labels')
                assert(all(isfield(valIn.labels,{'t' 'channels' ...
                    'channelUnits' 'channelNames'}))...
                    ,['The labels field, if present, must contain the'...
                    ' following fields: ''t'' ''channels''' ...
                    '''channelUnits'' and ''channelNames''.']);
                
                assert(...
                    isnumeric(valIn.labels.t) ...
                    && ~any(isnan(valIn.labels.t))...
                    && size(valIn.labels.t,2) == 1 ...
                    ,'t must be numeric and non-NaN.');
                
                assert(isequal(1 ...
                    ,numel(valIn.labels.channels)...
                    ,numel(valIn.labels.channelNames)...
                    ,numel(valIn.labels.channelUnits))...
                    ,['The channels channelNames and channelUnits'...
                    ' fields must have a signle element.'])
                
                nLabels = size(valIn.labels.t,1);
                assert(...
                    isequal(size(valIn.labels.channels{1}),[nLabels 1])...
                    ,['The channel must have the same number of'...
                    ' rows as the t vector']);
                
            end
            
            if isfield(valIn,'eyeTracking')
                assert(isa(valIn.eyeTracking,'EyeTrackingDataModel')...
                    ,['Use a EyeTrackingDataModel object as'...
                    ' the eyeTracking field.']);
            end
            
            obj.data = valIn;
        end
        
        
        %==================================================================
        function set.epochs(obj,valIn)
                        
            if isempty(valIn)
                obj.epochs = [];
                return
            end
            
            assert(isstruct(valIn) && isscalar(valIn) ...
                && isfield(valIn,'epochData')...
                ,['The epochs field must be a scalar stuct'...
                ' containing the epochData field.'])
            
            assert(istable(valIn.epochData)...
                ,'The ''epochData'' field must be a table.')
            
            varNames = valIn.epochData.Properties.VariableNames;
            assert(...
                ismember('startTime',varNames)...
                && ismember('endTime',varNames)...
                && ismember('epochName',varNames)...
                ,['The epochs field must contain the fields:'...
                ' ''startTime'', ''endTime'' and ''epochName''.'])
            
            assert(...
                isnumeric(valIn.epochData.startTime)...
                && isnumeric(valIn.epochData.endTime)...
                ,['The ''startTime'' and ''endTime'' variables'...
                ' must be numeric.']);
            
            assert(...
                iscellstr(valIn.epochData.epochName)...
                || iscategorical(valIn.epochData.epochName)...
                ,['The ''epochName'' var must be categorical or'...
                ' a cellstr.']);
            
            % Save:
            obj.epochs = valIn;
            
        end
        
        
        %==================================================================
        function set.physioDataInfo(obj,valIn)
            
            if isempty(valIn)
                obj.physioDataInfo = valIn;
                return
            end
            
            assert(isstruct(valIn) && isscalar(valIn) ...
                ,'physioDataInfo, if present, must be a scalar struct.')
            
            assert(all(isfield(valIn,{'sourceFilename' 'extracionUser' ...
                'extracionDate'}))...
                ,['physioDataInfo, if present, must contain the'...
                ' following fields: ''sourceFilename'''...
                '''extracionUser'' and' ...
                '''extracionDate'' [sic!].']);
            assert(...
                ischar(valIn.sourceFilename) ...
                && ischar(valIn.extracionUser) ...
                && ischar(valIn.extracionDate) ...
                ,'The physioDataInfo fields must be chars.')
           
            obj.physioDataInfo = valIn;
            
        end
        
        
        %==================================================================
        function genPhysioDataInfo(obj,fn)
            
            PDinfo.sourceFilename   = fn;
            PDinfo.extracionUser    = getenv('USERNAME');
            PDinfo.extracionDate    = datestr(now);
            obj.physioDataInfo      = PDinfo;
            
        end
        
        
        %==================================================================
        function saveToFile(obj,physioDataFilename,appendFlag)
            %saveToFile Saves the data inside the object to a PhysioData
            %file.
            %
            % Syntax:
            %
            %    saveToFile(obj,physioDataFilename[,appendFlag)]
            %
            %
            % Input/Output:
            %
            %    physioDataFilename - the output file name, w/o the
            %    .physioData Toolbox.
            %
            %    appendFlag - An optional parameter that causes the
            %    function to append the data inside an existing physioData,
            %    rather than overwritting everything. This leaves existing
            %    PhysioAnalyzer modules in tact, albeit with dirrent data.
            %
            %--------------------------------------------------------------
            
            assert(~isempty(obj.data),'Object contains no data to save.');
            
            % Parse inputs:
            if nargin < 3
                appendFlag = false;
            end
            
            % Get data:
            fileData.data = obj.data;
            
            % If there is eyeTracker data, reduce it to a struct:
            if isfield(fileData.data,'eyeTracking')
                fileData.data.eyeTracking ...
                    = returnStruct(fileData.data.eyeTracking);
            end
            
            % Get physioDataInfo:
            if ~isempty(obj.epochs)
                fileData.epochs = obj.epochs;
            end
            
            % Get physioDataInfo:
            if ~isempty(obj.physioDataInfo)
                fileData.physioDataInfo = obj.physioDataInfo; %#ok<STRNU>
            end
            
            % Construct save params:
            saveParams = {};
            
            % If the filename does not include the extension, add it:
            if length(physioDataFilename) > 11 ...
                    && strcmpi(physioDataFilename(end-10:end)...
                    ,'.physioData')
                newFileName = physioDataFilename;
            else
                newFileName = [physioDataFilename '.physioData'];
            end
            
            % Insert the append flag only if it was requested, and the file
            % exists:
            if ~(exist(newFileName...
                    ,'file') == 2)
                appendFlag = false;
            end
            if appendFlag
               saveParams = [saveParams '-append'];
            end
            
            % Save file:
            save(newFileName...
                ,'-struct','fileData',saveParams{:});
            
            % Feedback:
            if appendFlag
                fprintf('   > Appended file: %s.\n',newFileName);
            else
                fprintf('   > Created new file: %s.\n',newFileName);
            end
            
        end
        
        
        %==================================================================
        function varargout = loadSignalsFromAcq(obj,acqFilename,chans)
            %loadSignalsFromAcq Loads data from an acq file.
            %
            %   loadSignalsFromAcq(obj,acqFilename[,chans])
            %
            %--------------------------------------------------------------
            
            if nargin==2
                
                % Run get acq getter:
                [header, unit, fs, ~, marker, channelData] ...
                    = acqGet(acqFilename);
                
                %#function acqGetCB
                
            elseif nargin==3
                
                % Run get acq getter:
                [header, unit, fs, ~, marker, channelData] ...
                    = acqGet(acqFilename,chans);
            end
            
            % Add markers:
            if ~isempty(marker)
                dataTemp.labels.t            = vertcat(marker{:,2})/fs;
                dataTemp.labels.channels     = {vertcat(marker(:,3))};
                dataTemp.labels.channelUnits = {'[-]'};
                dataTemp.labels.channelNames = {'Labels'};
            end
            
            % Add info:
            physioDataInfoTemp.sourceFilename   = acqFilename;
            physioDataInfoTemp.extracionUser    = getenv('USERNAME');
            physioDataInfoTemp.extracionDate    = datestr(now);
            
            % Add data (note, only do this when channel data is available.
            % Because users can run the converter w/o any channels enabled,
            % the acqGet will run in metadataOnly mode, producing an empty
            % channelData, but returning all units and names, which will
            % crash the data modal):
            if ~isempty(channelData)
                dataTemp.signals.channels           = channelData;
                dataTemp.signals.channelUnits       = unit;
                dataTemp.signals.channelNames       = header;
                dataTemp.signals.channelDescription = strcat(...
                    {'Channel '''},header,{''' from acq.'});
                dataTemp.signals.fs                 = fs;
            end
            
            if nargout == 0
                % Save to object in its complete form:
                obj.data           = dataTemp;
                obj.physioDataInfo = physioDataInfoTemp;
            else
                varargout = {dataTemp physioDataInfoTemp};
            end
            
        end
        
        
        %==================================================================
        function s = struct(obj)
            % Overrides the struct(obj) function and returns a struct with
            % the raw data.
            
            % Copy props to struct fields:
            s.data           = obj.data;
            s.epochs         = obj.epochs;
            s.physioDataInfo = obj.physioDataInfo;
            
        end
    end
end