function [sigStructOut] = calcSignal(expr,sigStructIn)
% Uses an expression to generate calculation signals from numeric data.
%
%  Syntax:
%    [sigStructOut] = calcSignal(expr,sigStructIn);
%
%      where sigStructOut is the data generated by applying the expr to
%      sigStructIn.
%
%      Both sigStructOut and sigStructIn are data structs following the
%      PhysioData Toolbox convention; i.e., having the following fields:
%       # channels: 1xn cell with 1 signal vector (mx1 double ) per cell.
%       # channelNames: 1xn cell with the channel names.
%       # channelUnits: 1xn cell with the channel units.
%       # channelDescription: 1xn cell with the channel descriptions.
%       # fs: the sampling frequency.
%       # L: the number of samples (m in the mx1 signals).
%
%      expr is a string indicating the formula(s) to be used to generate
%      the new data. expr should start with a new channel name, followed by
%      its unit, enclosed in [ and ], then = and the expression. Muliple
%      calculations must be seperate by semicolon. Standard MATLAB syntax
%      must be used, and common functions, such as mean() and detrend() are
%      allowed. Note that all whitespaces will be removed from the
%      expression.
%
%      Example:
%        expr = 'ECG [mV] = (EXG7 - EXG8)/1000; EXT [mV] = EXG1/1000'
%
%        sigStructOut will then contain two channels: ECG and EXT, both
%        with the units mV. Those channels will be calculated by first
%        searching sigStructIn for the signals with the referenced names
%        (EXG7 and EXG8, and EXG1, respectively), and then using those
%        signals and the two formulas to generate the new data.
%
%
%  Elio Sjak-Shie, Sep 2019.
%--------------------------------------------------------------------------


%% Get new Headers:

% Assess complete expresion, and extract the individual channel headers and
% expressions:
exprParsed = regexp(strrep(expr,' ','')...
    ,'(^|;)(?<header>\D\w*)\[(?<unit>[^;]*)\]=(?<expr>[^;]*)','names');
sigStructOut.channelNames       = {exprParsed.header};
sigStructOut.channelDescription = {exprParsed.expr};
sigStructOut.channelUnits       = {exprParsed.unit};

% Add stuff to signals out:
sigStructOut.fs               = sigStructIn.fs;
sigStructOut.L                = sigStructIn.L;
    
% Return if no data was passed:
if nargin == 1
    return
end

% Run expressions:
sigStructOut.channels     = cell(1,numel(exprParsed));
sigStructOut.channels(:)  = {zeros(sigStructIn.L,1)};
for iExpr = 1 : numel(exprParsed)
    
    % Get the variables first (note that this should be the input argument
    % sequence of the function produced from the string expression):
    varsCell = symvar(exprParsed(iExpr).expr);
                
    % Run one of the methods for evaluating a string expression:            
    switch 'inline'
        %------------------------------------------------------------------
        case 'anon'
            % Use str2func to generate anaonymous function.
            %
            % Use this if the inline mode gets deprecated.
            %--------------------------------------------------------------
            
            % Generate an string represenration of an
            % anon function, then make the anon func:
            expFunc = str2func(...
                strcat('@(',strjoin(varsCell,','),') '...
                ,exprParsed(iExpr).expr));
            
            %--------------------------------------------------------------
        case 'inline'
            % Use fcnchk to generate an inline object, which is then
            % evaluated as an anon function.
            %
            % Altough inline objects are legacy, this appoach might be the
            % best since it vectorizes the expression and returns an error
            % message.
            %--------------------------------------------------------------
            
            % Make a vectorized inline object, which can then be evaluated:
            [expFunc, err] = fcnchk(exprParsed(iExpr).expr...
                ,varsCell{:},'vectorized') ; %#ok<DFCNCHK>
            assert(isempty(err),err.message);
            
            %--------------------------------------------------------------
        case 'symToolbox & muPad'
            % Use symbolic math. Might be slower and certainly does not
            % compile. Do not use!
            %--------------------------------------------------------------
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %NOTE: THIS DOESN'T WORK WHEN DEPLOYED!
            % > The Symolic Toolbox does not support compilation!
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            % Generate a symbolic expression:
            exprSym = evalin(symengine, exprParsed(iExpr).expr);
            
            % Turn the expression into a function (note that the input args
            % are in order stipulated in symVars):
            expFunc = matlabFunction(exprSym,'vars',varsCell);
            
    end
    
    % Find the vars in the headers:
    varIndxCel = cellfun(@(c) ...
        {find(strcmp(c,sigStructIn.channelNames))},varsCell);
    assert(all(cellfun(@numel,varIndxCel)==1)...
        ,'Error finding exactly one header named %s.'...
        ,listToString(varsCell(cellfun(@numel,varIndxCel)~=1)));
    
    % Eval expr:
    args = cellfun(@(ii) {sigStructIn.channels{ii}},varIndxCel);
    sigStructOut.channels{iExpr} = expFunc(args{:});
    
    
end

end





