function sOut = convert_BIOPAC(doWhat,setts)
% Converts BIOPAC files.
%
%  This converter contains no special settings.
%
%  See DESC.m for info.
%
%  Elio Sjak-Shie, June 2019
%--------------------------------------------------------------------------

%%

switch doWhat
    
    %----------------------------------------------------------------------
    case 'getConverterInfo'
        % Returns info about the converter.
        %------------------------------------------------------------------
        
        % Include a description of the converter
        sOut.name   = 'BIOPAC';
        sOut.desc = ['Converts BIOPAC AcqKnowledge '...
            '.acq and .mat files'];
        
        % Make a struct for the customSettings:
        sOut.setupSetts.biopacExt.type ...
            = 'choice';
        sOut.setupSetts.biopacExt.value ...
            = 1;
        sOut.setupSetts.biopacExt.label ...
            = 'AcqKnowledge Output Type';
        sOut.setupSetts.biopacExt.desc ...
            = ['The BIOPAC converter supports'...
            ' AcqKnowledge (v3.9 - v5.0.2) data saved'...
            ' as .acq files or exported as .mat files.'...
            newline newline ...
            'Because of a limitation of the BIOPAC File API, when'...
            ' converting .acq files,'...
            ' all channels must have the same sample-rate'...
            ' and sample-count.'...
            ' This limitation does not hold for data exported as .mat'...
            ' files. This format, however, does not include any'...
            ' labels or events.'...
            ];
        sOut.setupSetts.biopacExt.choices ...
            = {...
            '<html>Convert AcqKnowledge <b>.acq</b> files.<html>'...
            '<html>Convert AcqKnowledge <b>.mat</b> files.<html>'};
        
        
        %------------------------------------------------------------------
    case 'setupConverter'
        %------------------------------------------------------------------
        
        % Return the extension:
        if setts.setupSetts.biopacExt.value == 1
            sOut.ext    = 'acq';
        else
            sOut.ext    = 'mat';
        end
        
        return
        
        % Make a struct for the customSettings:
        sOut.customSetts.choiceOnes.type ...
            = 'choice';
        sOut.customSetts.choiceOnes.value ...
            = 1;
        sOut.customSetts.choiceOnes.label ...
            = 'What to do';
        sOut.customSetts.choiceOnes.desc ...
            = ['This is a very long description, which the'...
            ' file converter will probably need to wrap'...
            ' quite a few times in order for it to fit inside the'...
            ' BlurBox.'...
            newline newline ...
            'It even features a two newline characters.'];
        sOut.customSetts.choiceOnes.choices ...
            = {...
            'Do thing 1.'...
            'Do thing 2, which has a much longer description.'...
            'Do that other thing you do so well.'...
            };
        
        % Make a struct for the customSettings:
        sOut.customSetts.checkOneTwo.type ...
            = 'check';
        sOut.customSetts.checkOneTwo.value ...
            = [false true];
        sOut.customSetts.checkOneTwo.label ...
            = 'Check yo self';
        sOut.customSetts.checkOneTwo.desc ...
            = 'Check the checkboxes below, if you like.';
        sOut.customSetts.checkOneTwo.choices ...
            = {...
            'Option numero uno.'...
            'Option numero dos.'...
            };
        
        % Make a struct for the customSettings:
        sOut.customSetts.labelCols.type ...
            = 'text';
        sOut.customSetts.labelCols.value ...
            = '';
        sOut.customSetts.labelCols.label ...
            = 'Label Generation';
        sOut.customSetts.labelCols.desc ...
            = ['The Toolbox can automatically generate labels'...
            ' based on metadata columns in the gazedata files.'...
            ' bla bla bla.'];   
        
        
        %------------------------------------------------------------------
    case 'getFileInfo'
        % Returns metadata about a file.
        %
        % Note, exceptions are thrown when errors are encountered. The File
        % Converter will catch and handle them as necessary.
        %------------------------------------------------------------------
        
        % Get filename:
        rawFileName      = setts.rawFileName;
        [~,~,rawFileExt] = fileparts(rawFileName);
        switch rawFileExt
            case '.acq'
                
                % Test acqget.dll, which is a common source of issues:
                assert(exist('acqfile.dll','file')==2 ...
                    ,'The acqfile.dll is missing.');
                assert(exist(rawFileName,'file')==2 ...
                    ,'The raw file is missing.');
                s = dir(which('acqfile.dll'));
                assert(~isempty(s) && isequal(s.bytes,94208)...
                    ,'The acqfile.dll is corrupt.');
                
                % Run converter function:
                [signals.channelNames, signals.channelUnits, signals.fs...
                    , signals.L...
                    , marker, ~] ...
                    = acqGet(rawFileName,[]);
                
                % Make output struct:
                sOut.signals   = signals;
                
                % Add markers:
                if ~isempty(marker)
                    sOut.labels.t            = vertcat(marker{:,2})/signals.fs;
                    sOut.labels.channels     = {vertcat(marker(:,3))};
                    sOut.labels.channelUnits = {'[-]'};
                    sOut.labels.channelNames = {'Labels'};
                    
                    % The above is not necessary, and may be removed. Only
                    % the count is used by the file converter.
                    sOut.labels.count        = numel(sOut.labels.t);
                    
                end
            case '.mat'
                
                % Load data:
                matData = load(rawFileName);
                assert(strcmp(matData.isi_units,'ms')...
                    ,'Only ''ms'' ''isi_units'' currently supported.');
                
                % Parse data to output:
                sOut.signals = makeSig(rawFileName,false);
                
            otherwise
                error('Unexpected extension.')
        end
        
        
        %------------------------------------------------------------------
    case {'convertFile' 'getData'}
        % In either the convertFile or getData cases, get the data first,
        % then either save it or return it.
        %------------------------------------------------------------------
        
        % Make a PhysioDataFileModel object:
        physDatObj = PhysioDataFileModel();
        
        % Load data:
        [~,~,rawFileExt] = fileparts(setts.rawFileName);
        switch rawFileExt
            case '.acq'
                
                % Use the data model to get the data from an acq file:
                physDatObj.loadSignalsFromAcq(setts.rawFileName...
                    ,setts.signalsToExtract - 1);
                
            case '.mat'
                
                % Get the data from a mat file:
                physDatObj.data.signals = makeSig(setts.rawFileName,true);
                physDatObj.genPhysioDataInfo(setts.rawFileName);
                
        end
        
        % Process data:
        switch doWhat
            case 'convertFile'
                physDatObj.saveToFile(setts.newFileName,setts.doAppend);
            case 'getData'
                sOut = struct(physDatObj);
        end
        
    otherwise
        error('Wrong flag')
end


end


%==========================================================================
function signals = makeSig(rawFileName,doData)
% Loads a AcqKnowledge MAT file into a struct.

% Load data:
matData = load(rawFileName);
assert(strcmp(matData.isi_units,'ms')...
    ,'Only ''ms'' ''isi_units'' currently supported.');

% Parse data to output:
signals.channelNames = strtrim(cellstr(matData.labels))';
signals.channelUnits = strtrim(cellstr(matData.units))';
signals.fs           = 1/(matData.isi/1000);
signals.L            = size(matData.data,1);
signals.channelDescription = strcat(signals.channelNames...
    ,{' ['},signals.channelUnits,{']'});

if doData
    signals.channels     = num2cell(matData.data,1);
end

end


