function varargout = convert_BIOSEMI(doWhat,setts)
% Converts BIOPAC files.
%
%  This converter contains no special settings.
%
%  See DESC.m for info.
%
%  Elio Sjak-Shie, June 2019
%--------------------------------------------------------------------------

%%

switch doWhat
    
    %----------------------------------------------------------------------
    case 'getConverterInfo'
        % Returns info about the converter.
        %------------------------------------------------------------------
        
        % Include a description of the converter
        sOut.name = 'Biosemi';
        sOut.desc = 'Converts Biosemi .bdf files';
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
        %------------------------------------------------------------------
    case 'setupConverter'
        %------------------------------------------------------------------
        
        
        % Return the extension:
        sOut.ext    = 'bdf';
        
        % Make a struct for the customSettings:
        sOut.customSetts.sigExpr.type ...
            = 'text';
        sOut.customSetts.sigExpr.value ...
            = 'ECG [mV] = detrend(EXG7 - EXG8)/1000';
        sOut.customSetts.sigExpr.label ...
            = 'Calculation Channels';
        sOut.customSetts.sigExpr.desc ...
            = ['Use this field to generate new data from signals'...
            ' in the file. Refer to one or more signals by their labels'...
            ', and perform the necessary calculation to'...
            ' calculate new data.'...
             newline newline...
             'Start with a new channel name,'...
            ' followed by its unit enclosed in [ and ], then = and'...
            ' the expression. Seperate multiple'...
            ' calculations with semicolons, and use standard MATLAB'...
            ' syntax (common functions, such as mean()'...
            ' and detrend(), are allowed). Note that'...
            ' all whitespaces will be removed from the expression'...
            ' , and that the'...
            ' BIOSEMI units are uV by default.'...
            newline newline ...
            'Example: ECG [mV] = (EXG7 - EXG8)/1000;'...
            ' EXT [mV] = EXG1/1000'...
            ];
        
        % Make a struct for the customSettings:
        sOut.customSetts.convertMarkButton.type ...
            = 'check';
        sOut.customSetts.convertMarkButton.value ...
            = [true true];
        sOut.customSetts.convertMarkButton.label ...
            = 'Markers and Button Presses';
        sOut.customSetts.convertMarkButton.desc ...
            = ['Enable to extract markers and button-presses'...
            ' from a channel named ''Status''. The Converter assumes'...
            ' that the first 8 bits encode the markers, which will be'...
            ' converted into a marker channel; and the subsequent'...
            ' 8 bits represent button presses, which will be'...
            ' converted to labels. Button presses are not debounced.'];
        sOut.customSetts.convertMarkButton.choices ...
            = {...
            'Convert markers.'...
            'Convert button-presses.'...
            };
        
        varargout{1} = sOut;
        return
        
        
        %------------------------------------------------------------------
    case {'getFileInfo'}
        %------------------------------------------------------------------
        
        % get only the metadata:
        try
            [~,data.signals] = read_bdf(setts.rawFileName,false);
        catch ball
            error('Error generating data: %s',ball.message);
        end
        
        varargout{1} = data;
                
        
        %------------------------------------------------------------------
    case {'convertFile' 'getData'}
        %------------------------------------------------------------------
        
        % Get the data:
        data = processBiosemiData(setts);
        
        % Make a PhysioDataFileModel object:
        physDatObj = PhysioDataFileModel();
        physDatObj.data = data;
        physDatObj.genPhysioDataInfo(setts.rawFileName);
        
        % Process data:
        switch doWhat
            case 'convertFile'
                physDatObj.saveToFile(setts.newFileName...
                    ,setts.doAppend);
                return
            case 'getData'
                varargout{1} = struct(physDatObj);
                return
        end
        
        
    otherwise
        error('Wrong flag')
end


end


%==========================================================================
function data = processBiosemiData(setts)
% Gets data frpom a bdf  file, then generates the marker and button-label
% data.

% Parse inputs:
rawFileName = setts.rawFileName;
convertMark   = setts.customSetts.convertMarkButton.value(1);
convertButons = setts.customSetts.convertMarkButton.value(2);
calcExpr      = setts.customSetts.sigExpr.value;

% Get the data from the file:
[~,bdfSignals] = read_bdf(rawFileName,true);
        
% Process the expression (to determine the calculation channels):
calcSignals = calcSignal(calcExpr,bdfSignals);

% Merge (only keep the requested channels from the bdf):
for F = fieldnames(calcSignals)'
    if ~isscalar(bdfSignals.(F{1}))
        data.signals.(F{1}) =...
            [bdfSignals.(F{1})(setts.signalsToExtract) calcSignals.(F{1})];
    else
        data.signals.(F{1}) = calcSignals.(F{1});
    end
end

% Get markers and button presses:
nChans       = numel(data.signals.channelNames);
if convertMark || convertButons
    markChan = strcmp(bdfSignals.channelNames,'Status');
    assert(sum(markChan)==1 ...
        ,'Not exactly one Status channel found.');
    mrkInt8x4 = typecast(uint32(bdfSignals.channels{markChan}),'uint8');
end
if convertMark
    
    data.signals.channelUnits(nChans+1) = {'-'};
    data.signals.channelNames(nChans+1) = {'markers'};
    data.signals.channelDescription(nChans+1) ...
        = {'Biosemi markers'};
    data.signals.channels(nChans+1) ...
        = {double(mrkInt8x4(1:4:end))};
end

if convertButons
    
    % Get the buttons signal, and make it bin:
    buttons   = double(mrkInt8x4(2:4:end));
    buttBin   = de2bi(buttons,8);
    
    % Function to find button presses and releases:
    findPressRelease = @(bitNum) ...
        deal(find(diff([0;buttBin(:,bitNum)])==1)...
        ,find(diff([0;buttBin(:,bitNum)])==-1));
    
    % Find button clicks and releases, per bit:
    bitNums = 1:8;
    [presses,releases] ...
        = arrayfun(@(bitNum) findPressRelease(bitNum)...
        ,bitNums,'UniformOutput',false);
    [presLabels,releaseLabels] = arrayfun(...
        @(bitNum) deal(...
        repmat({sprintf('Button %i Pressed',bitNum)}...
        ,numel(presses{bitNum}),1)...
        ,repmat({sprintf('Button %i Released',bitNum)}...
        ,numel(releases{bitNum}),1))...
        ,bitNums,'UniformOutput',false);
    
    % Make labels:
    label_t          = vertcat(presses{:},releases{:})/bdfSignals.fs;
    label_name       = vertcat(presLabels{:},releaseLabels{:});
    [label_t,sortII] = sort(label_t);
    label_name       = label_name(sortII);
    
    data.labels.t            = label_t;
    data.labels.channels     = {label_name};
    data.labels.channelUnits = {'[-]'};
    data.labels.channelNames = {'Labels'};
    
end




end


