function varargout = convert_EYELINK(doWhat,setts)
% Converts gazedata files with the gazedata extention.
%
%  See DESC.m for info.
%
%  Elio Sjak-Shie, June 2019
%
% Eyelink data info:
%  > https://github.com/iandol/opticka/blob/master/communication/edfmex.m
%
%  > https://github.com/tz-lom/eyelinkReader/blob/master/src/edfapi/headers/edf_data.h
%
%--------------------------------------------------------------------------


%%

switch doWhat
    
    %----------------------------------------------------------------------
    case 'getConverterInfo'
        % Returns info about the converter.
        %------------------------------------------------------------------
        
        
        % Include a description of the converter
        sOut.name = 'EyeLink';
        sOut.desc = 'Converts SR Research''s EyeLink .edf files';
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
        %------------------------------------------------------------------
    case 'setupConverter'
        % Returns settings for the convertion.
        %------------------------------------------------------------------
        
        % Determine extension:
        sOut.ext    = 'edf';
        
        % Settings:
        sOut.customSetts.remSysMsgs.type ...
            = 'text';
        sOut.customSetts.remSysMsgs.value ...
            = ['(!.*)|(DRIFTCORRECT.*)|(DISPLAY_COORDS.*)|(RECCFG.*)'...
            '|(ELCL.*)|(GAZE_COORDS.*)|(THRESHOLDS.*)|(CAMERA_LENS.*)'...
            '|(PUPIL_DATA_TYPE.*)|(VALIDATE.*)'];
        sOut.customSetts.remSysMsgs.label ...
            = 'Remove EyeLink System Events';
        sOut.customSetts.remSysMsgs.desc ...
            = ['The File Converter automatically converts EyeLink'...
            ' messages to eye-tracking events that can be used in the'...
            ' PhysioData Toolbox.'...
            ' The EyeLink data, however, contains many system events' ...
            ' that are not essential for data analysis.'...
            newline newline...
            'If desired, and to prevent clutter, the File converter'...
            ' can remove events that match the regular expression'...
            ' below. The default value of the expression removes'...
            ' system events, except errors. Leave the field blank to'...
            ' convert all messages to eye-tracking events.'];
        
        % Make a struct for the customSettings:
        sOut.customSetts.eyelinkEprimeSetts.type ...
            = 'check';
        sOut.customSetts.eyelinkEprimeSetts.value ...
            = false;
        sOut.customSetts.eyelinkEprimeSetts.label ...
            = 'EyeLink via E-Prime Options';
        sOut.customSetts.eyelinkEprimeSetts.desc ...
            = ['If the experiment was conducted using the'...
            ' approach outlined in the standard E-Prime EyeLink'...
            ' example,'...
            ' the converter can automatically correct for onset delays.'...
            ' This involves'...
            ' subtracting the offset encoded in "onset-messages"'...
            ' from that message''s timestamp.'...
            ' The offset is then also removed from the message string.'];
        sOut.customSetts.eyelinkEprimeSetts.choices ...
            = {...
            'Correct "<offset>  <Stim>_Onset" messages.'...
            };
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
        %------------------------------------------------------------------
    case 'getFileInfo'
        % Returns metadata about a file.
        %
        % Note, exceptions are thrown when errors are encountered. The File
        % Converter will catch and handle them as necessary.
        %------------------------------------------------------------------
        
        % Get filename:
        rawFileName = setts.rawFileName;
        
        % Read raw edf file:
        doSamples = true;
        if doSamples
            
            % https://github.com/iandol/opticka/blob/master/communication/edfmex.m
            rawEDF          = edfmex(rawFileName,0,0,{'time' 'pa'},1);
            
            % Obtain the time stamps from the raw data
            t_ms            = double(rawEDF.FSAMPLE.time)';
            
            % Obtain the zero time stamp
            zeroTime_ms     = t_ms(1);
            
            % Calculate the time stamps starting from timepoint 0
            t_ms            = (t_ms - zeroTime_ms);
            
            % Save pupil data in diameter struct:
            diameter.t_ms  = t_ms;
            diameter.L     = double(rawEDF.FSAMPLE.pa(1,:)');
            diameter.R     = double(rawEDF.FSAMPLE.pa(2,:)');
            
            % Delete rows with no data:
            noDataRows = (diameter.L <= 0) & (diameter.R <= 0);
            diameter.t_ms(noDataRows)  = [];
            diameter.L(noDataRows)     = [];
            diameter.R(noDataRows)     = [];
            
            % Construct info struct:
            sOut.eyeTracking.nL           = sum(diameter.L > 0);
            sOut.eyeTracking.nR           = sum(diameter.R > 0);
            sOut.eyeTracking.nLR          = numel(diameter.t_ms);
            
        else
            rawEDF          = edfmex(rawFileName,0,0,0,1);
            sOut.eyeTracking.nL           = NaN;
            sOut.eyeTracking.nR           = NaN;
            sOut.eyeTracking.nLR          = NaN;
        end
        sOut.eyeTracking.elapsedTime ...
            = diff(double([rawEDF.FEVENT([1 end]).sttime]))/1000;
        sOut.eyeTracking.diamFs      ...
            = double(mode([rawEDF.RECORDINGS.sample_rate]));
        
        % ET: diff(double([rawEDF.FEVENT([1 end]).sttime]))/1000
        
        % Set the out and return:
        varargout{1} = sOut;
        
        return
        
        
        %------------------------------------------------------------------
    case {'convertFile' 'getData'}
        % In the convertFile case, get the data first and save it as a
        % PhysioData file.
        %
        %------------------------------------------------------------------
        
        % Call data getter function:
        physDatObj = makePhysDataObj(setts);
        physDatObj.genPhysioDataInfo(setts.rawFileName);
        
        % Save to file, or return data:
        switch doWhat
            case 'convertFile'
                saveToFile(physDatObj,setts.newFileName,setts.doAppend);
                return
            case 'getData'
                varargout{1} = struct(physDatObj);
                return     
        end
        
        
        %------------------------------------------------------------------
    otherwise
        error('Wrong flag')
end


end


%==========================================================================
function physDatObj = makePhysDataObj(setts)
% Reads an Eyelink edf file and return its data in a PhysioDataFileModel
% instance.

% Read raw edf file:
rawEDF          = edfmex(setts.rawFileName);

% Obtain the time stamps from the raw data
t_ms            = double(rawEDF.FSAMPLE.time)';

% Obtain the zero time stamp (placed at the first message (probably a
% system message indicating the display coords or refresh rate)):
zeroTime_ms     = double(rawEDF.FEVENT(1).sttime);

% Calculate the time stamps starting from timepoint 0
t_ms            = (t_ms - zeroTime_ms);

% Save pupil data in diameter struct:
diameter.t_ms  = t_ms;
diameter.L     = double(rawEDF.FSAMPLE.pa(1,:)');
diameter.R     = double(rawEDF.FSAMPLE.pa(2,:)');

% Delete rows with no data:
diameter.L(diameter.L <= 0)  = NaN;
diameter.R(diameter.R <= 0)  = NaN;
noDataRows = isnan(diameter.L) & isnan(diameter.R);
diameter.t_ms(noDataRows)  = [];
diameter.L(noDataRows)     = [];
diameter.R(noDataRows)     = [];

% If an eye has no data, delete it:
if all(isnan(diameter.L))
    diameter.L = [];
end
if all(isnan(diameter.R))
    diameter.R = [];
end


%% Process Events (Messages):

% Process events (only extract relevant, nonempty events):
evtRows = arrayfun(@(f) ~isempty(f.message),rawEDF.FEVENT);
eventData.t_ms = ...
    double(...
    vertcat(rawEDF.FEVENT(evtRows).sttime)...
    )...
    - zeroTime_ms;
eventData.name = {rawEDF.FEVENT(evtRows).message}';

% This converter includes the option to remove Eyelink system messages,
% which are not always essential for analysis, and may clog the data. This
% is done using a user-editable expression, which defines the system
% events. These conforming events are then removed:
try
    if ~isempty(setts.customSetts.remSysMsgs.value)
        
        % Find the system events using the expression, then delete them:
        sysMsgs = regexp(eventData.name...
            ,setts.customSetts.remSysMsgs.value, 'match','once');
        sysMsgsIndx  = ~cellfun(@isempty,sysMsgs);
        eventData.t_ms(sysMsgsIndx) = [];
        eventData.name(sysMsgsIndx) = [];
        
    end
catch
    error('Cannot process system message expression.');
end

% This converter also allows users to correct for delays when using the
% standard E-Prime & Eyelink approach (from the Eyelink example). This
% consists of encoding the delay in an onset indicating message. The code
% below removes this onset from the message body, and uses is to correct
% the timestamp:
try
    if setts.customSetts.eyelinkEprimeSetts.value
        
        % Onset messages:
        onsetMsgsAll = regexp(eventData.name...
            ,'(?<offset>\d+)\s{0,2}(?<name>.*_Onset)', 'names','once');
        onsetMsgRows = ~cellfun(@isempty,onsetMsgsAll);
        onsetMsgs = [onsetMsgsAll{onsetMsgRows}];
        
        % Correct the times:
        offsets = cellfun(@str2double,{onsetMsgs.offset}');
        eventData.t_ms(onsetMsgRows) ...
            = eventData.t_ms(onsetMsgRows) - offsets;
        
        % Remove offset from names:
        eventData.name(onsetMsgRows) = {onsetMsgs.name}';
        
    end
catch
    error('Cannot process offsets.');
end

% Trim all messages:
eventData.name = strtrim(eventData.name);

% Sort labels:
assert(issorted(eventData.t_ms),['Messages are not'...
    ' sorted after offset correction.']);


%% Make EyeTrackingDataModel Object:

% Make an EyeTrackingDataModel object:
ETobj = EyeTrackingDataModel();

% Fill in general info ET data:
ETobj.filename     = setts.rawFileName;
[~,ETobj.name,~]   = fileparts(setts.rawFileName);
ETobj.diameterUnit = 'px';

ETobj.raw_t_ms_max = max(t_ms);

% Store diameter struct in ETobj:
ETobj.diameter = diameter;

% Store labels struct in ETobj:
ETobj.labels = eventData;

% If requested, include a copy of the raw data:
if setts.includeRawCopy
    
    % Also store the raw values for use later:
    ETobj.UserData.rawData = rawEDF;
    
    % Also store the zeroTime_ms:
    ETobj.UserData.zeroTime_ms = zeroTime_ms;
    
end

% Fix empty gazedata issue:
ETobj.gazeData = [];

% Make file obj:
physDatObj = PhysioDataFileModel();
physDatObj.data.eyeTracking = ETobj;


end

