function varargout = convert_GAZEDATA(doWhat,setts)
% Converts gazedata files with the gazedata extention.
%
%  See DESC.m for info.
%
%  Elio Sjak-Shie, June 2019
%--------------------------------------------------------------------------

% Names of the file types:
gazedatastr = 'EET files with .gazedata extension.';
textstr     = 'EET files with .txt extension.';


%%

switch doWhat
    
    %----------------------------------------------------------------------
    case 'getConverterInfo'
        % Returns info about the converter.
        %------------------------------------------------------------------
        
        
        % Include a description of the converter
        sOut.name = 'EET Output';
        sOut.desc = 'Converts EET 2.x - 3.2 files';
        
        % Make a struct for the customSettings:
        sOut.setupSetts.gazedataExt.type ...
            = 'choice';
        sOut.setupSetts.gazedataExt.value ...
            = 1;
        sOut.setupSetts.gazedataExt.label ...
            = 'EET Output Type';
        sOut.setupSetts.gazedataExt.desc ...
            = ['The EET converter supports files with either the'...
            ' ''.gazedata'' extension (EET 2.x - 3.1), or the'...
            ' ''.txt'' extension (EET 3.2).' newline newline ...
            'Please select the type'...
            ' you wish to convert.'];
        sOut.setupSetts.gazedataExt.choices ...
            = {gazedatastr textstr};
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
        %------------------------------------------------------------------
    case 'setupConverter'
        % Returns settings for the convertion.
        %------------------------------------------------------------------
        
        % Determine extension:
        if setts.setupSetts.gazedataExt.value == 1
            sOut.ext    = 'gazedata';
        else
            sOut.ext    = 'txt';
        end
        
        % Make a struct for the customSettings:
        sOut.customSetts.labelCols.type ...
            = 'text';
        sOut.customSetts.labelCols.value ...
            = 'CurrentObject';
        sOut.customSetts.labelCols.label ...
            = 'Eye-Tracking Event Generation';
        sOut.customSetts.labelCols.desc ...
            = ['Eye-tracking events can automatically'...
            ' be generated by finding'...
            ' the start and end of each contiguous section of values'...
            ', or a combination of values, in one or more'...
            ' metadata columns. Specify the'...
            ' name of the column(s) here; seperate multiple columns with'...
            ' semicolons. Leave blank to disable.'];
        sOut.customSetts.labelCols.desc2 ...
            = ['The File Converter can automatically generate eye-tracking events'...
            ' based on entries in one or more metadata columns in the'...
            ' gazedata files. These'...
            ' events can then be referenced in the Toolbox'...
            ' when defining epoch.' newline newline...
            'The events are generated by finding the start and'...
            ' end of each contiguous section of entries in the columns'...
            ' specified here. Start and end'...
            ' events are then created at those points.'...
            ' Multiple columns can be specified by seperating them'...
            ' with semicolons. In that case, the values of the columns'...
            ' are first combined into a single entry through'...
            ' concatenation,'...
            ' then analyzed as mentioned above.'];
        
        sOut.customSetts.aoiCols.type ...
            = 'text';
        sOut.customSetts.aoiCols.value ...
            = '';
        sOut.customSetts.aoiCols.label ...
            = 'AOI Analysis';
        sOut.customSetts.aoiCols.desc ...
            = ['Area of Interest (AoI) hit data can be exported for'...
            ' analysis in the Toolbox if the gazedata file contains a'...
            ' single column where each row indicates the current'...
            ' AoI (if there is one). Enter the name of the AoI column'...
            ' to activate (e.g. AOIStimulus), leave empty to disable.'];
        
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
        %------------------------------------------------------------------
    case 'getFileInfo'
        % Returns metadata about a file.
        %
        % Note, exceptions are thrown when errors are encountered. The File
        % Converter will catch and handle them as necessary.
        %------------------------------------------------------------------
        
        % Read data:
        [dataTable, diamData, ~] = read_EET(setts.rawFileName,true);
        
        % Construct info struct:
        sOut.eyeTracking.nL          = sum(~isnan(diamData.L));
        sOut.eyeTracking.nR          = sum(~isnan(diamData.R));
        sOut.eyeTracking.nLR         = sum(~isnan(diamData.t_ms));
        sOut.eyeTracking.elapsedTime = dataTable.t_ms(end)/1000;
        sOut.eyeTracking.diamFs      = 1/median(diff(diamData.t_ms)./1000);
        
        % Set the out and return:
        varargout{1} = sOut;
        
        return
        
        
        %------------------------------------------------------------------
    case 'convertFile'
        % In the convertFile case, get the data first and save it as a
        % PhysioData file.
        %
        %------------------------------------------------------------------
        
        % Parse inputs:
        physDatObj = makePhysDataObj(setts);
        saveToFile(physDatObj,setts.newFileName,setts.doAppend);
        
        % Return, no output necessary:
        return
        
        
        %------------------------------------------------------------------
    case 'getData'
        % Gets and returns the data.
        %------------------------------------------------------------------
        
        % Make a PhysioDataFileModel object and return data:
        physDatObj = makePhysDataObj(setts);
        varargout{1} = struct(physDatObj);
        return
        
    otherwise
        error('Wrong flag')
end


end


%==========================================================================
function physDatObj = makePhysDataObj(setts)

% Read data:
[dataTable, diamData, info] = read_EET(setts.rawFileName,false);

% Make a PhysioDataFileModel and ET objects:
etObj      = EyeTrackingDataModel();
physDatObj = PhysioDataFileModel();

% Store data:
etObj.diameter = diamData;

% Process labels:
if ~isempty(setts.customSetts.labelCols.value)
    
    % Get labels:
    headers = strsplit(setts.customSetts.labelCols.value, ';');
    foundHeaders = ismember(headers, dataTable.Properties.VariableNames);
    assert(all(foundHeaders)...
        , 'The requested label column%s (%s) could not be found.'...
        , ifFcn(sum(~foundHeaders)==1,'','s')...
        , listToString(headers(~foundHeaders)));
    labelsCell = table2cell(dataTable(:,headers));
    labels = strtrim(arrayfun(@(ii) {strjoin(labelsCell(ii,:))} ...
        ,(1:size(labelsCell,1))'));
    [labels,eventSections] = analyzeList(...
        dataTable.t_ms,labels);
    
    % Store data:
    etObj.labels        = labels;
    etObj.raw_t_ms_max  = info.max_t_ms;
    etObj.eventSections = eventSections;
    
    clearvars headers labelsCell
    
end

% Process AoIs:
if ~isempty(setts.customSetts.aoiCols.value)
    
    aoiHeader = setts.customSetts.aoiCols.value;
    assert(...
        ismember(...
        aoiHeader,dataTable.Properties.VariableNames)...
        ,'The requested AoI column ("%s") could not be found.'...
        ,aoiHeader);
    labelsCell = table2cell(dataTable(:,aoiHeader));
    
    % Remove alt. blanks (e.g. LvD dataset has '_'):
    blanks = {'_' ' ' '.'};
    labelsCell(ismember(labelsCell,blanks)) = {''};
    
    % Make the hit struct and put it in the ET object:
    [~, eventSections] = analyzeList(dataTable.t_ms,labelsCell);
    
    % Make the gazedata struct and add it to the ET object (the Width and
    % Height are unimportant):
    gazeData = struct(...
        'snapshotName','EXP'...
        ,'snapshotWidth',800 ...
        ,'snapshotHeight',600);
    gazeData.RoI = eventSections;
    etObj.gazeData = gazeData;
    
end


% Make PhysiOData object:
[~,etObj.name,~]            =  fileparts(setts.rawFileName);
physDatObj.data.eyeTracking = etObj;
physDatObj.genPhysioDataInfo(setts.rawFileName);

% If requested, include a copy of the raw data:
if setts.includeRawCopy
    etObj.UserData.dataTable    = dataTable;
    etObj.UserData.info         = info;
end


end

