function varargout = convert_VUAMS(doWhat,setts)
% Converts VUAMS 5FS files.
%
% Elio Sjak-Shie, June 2019.
%--------------------------------------------------------------------------

%%

switch doWhat
    
    case 'getConverterInfo'
        % Returns converter info.
        
        % Include a description of the converter
        sOut.name   = 'VU-AMS';
        sOut.desc   = 'Converts VU-AMS 5FS files';
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
    case 'setupConverter'
        
        % Return the extension (no special options here):
        sOut.ext    = '5fs';
                
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
    case 'getFileInfo'
        % Returns metadata info about a single 5FS file.
        
        % Get filename:
        rawFileName = setts.rawFileName;
        assert(exist(rawFileName,'file')==2 ...
            ,'The raw file is missing.');
        
        % Get the data:
        dataRaw = getData(rawFileName);
        
        % Make output struct:
        sOut.signals   = rmfield(dataRaw.signals,'channels');
        sOut.signals.L = numel(dataRaw.signals.channels{1});
        
        % Add markers:
        if isfield(dataRaw,'labels')
            
            % Only the count is used by the file converter, but lets leave
            % the labels in anyways:
            sOut.labels       = dataRaw.labels;
            sOut.labels.count = numel(sOut.labels.t);
            
        end
        
        % Set the out and return:
        varargout{1} = sOut;
        return
        
        
    case {'convertFile' 'getData'}
        % Reads a single 5fs file and converts it.
        
        
        % Make a PhysioDataFileModel object:
        physDatObj = PhysioDataFileModel();
        
        % Get the data:
        dataRaw = getData(setts.rawFileName);
        
        % Remove unnecessary channels:
        chanNums = setts.signalsToExtract;
        assert(isfield(dataRaw,'signals')...
            ,'No data in file.');
        assert(...
            max(chanNums) <= numel(dataRaw.signals.channels)...
            && max(chanNums) <= numel(dataRaw.signals.channelUnits)...
            && max(chanNums) <= numel(dataRaw.signals.channelNames)...
            ...
            && max(chanNums) ...
            <= numel(dataRaw.signals.channelDescription)...
            ,'User requested channel that does not exist.');
        
        % Copy requested data:
        data = dataRaw;
        sigs.fs                 = dataRaw.signals.fs;
        sigs.channels           = dataRaw.signals.channels(chanNums);
        sigs.channelUnits       = dataRaw.signals.channelUnits(chanNums);
        sigs.channelNames       = dataRaw.signals.channelNames(chanNums);
        sigs.channelDescription ...
            = dataRaw.signals.channelDescription(chanNums);
        data.signals = sigs;
        
        % Insert data:
        physDatObj.data = data;
        physDatObj.genPhysioDataInfo(setts.rawFileName);
        
        % Save file:
        switch doWhat
            case 'convertFile'
                
                saveToFile(physDatObj,setts.newFileName,setts.doAppend);
                
                % Return, no output necessary:
                return
                
            case 'getData'
                
                % return the data:
                varargout{1} = struct(physDatObj);
                return
                
        end
        
        
        
    otherwise
        error('wrong flag')
end

end



%% Helper Functions:

function data = getData(fileName)
% Reads VU-AMS data by calling the read5fs function.

        try
            data = read5fs(fileName);
        catch ME
            
            error('read5fs crashed, probably an unsupported file type.');
            
        end
         
end

