function dZdt = getDZDT(Z0, diffFIR_order, diffFIR_HPCF, diffFIR_LPCF, fs)
%% Explanation
% This function calculates dZ/dt from Z0
% -------------------------------------------------------------------------
%
% Elio Sjak-Shie and Iris Spruit, SOLO, 2019
% Function that calculates dZ/dt from a Z0 signal
%
% Input:
% - Z0 = the Z0 signal (n x 1 double)
% - diffFIR_order = filter order
% - diffFIR_HPCF = passband frequency
% - diffFIR_LPCF = stopband frequency
% - fs = sampling rate
%
% Output:
% - DZDT signal
%
%
% -------------------------------------------------------------------------


%%

% Get signal:
sig           = Z0;

% Use a differentiator FIR filter to reduce noise when calculating
% the higher derivatives:
diffFilt = designfilt(...
    'differentiatorfir'...
    ,'FilterOrder',diffFIR_order ...
    ,'PassbandFrequency',diffFIR_HPCF ...
    ,'StopbandFrequency',diffFIR_LPCF ...
    ,'SampleRate',fs);
 
% Calculate the filter delay:
diffFiltDelaySamples = mean(grpdelay(diffFilt));
 
% Calculate zscored dZdt_d from the zero-appended dZdt using the
% diff-FIR, and compensate for the filter delay:
dZdt = -fs * filter(diffFilt, [sig; zeros(diffFiltDelaySamples,1)]);

% Remove the artifacts caused by the filter at the start and end of the
% signal:
dZdt((end-diffFiltDelaySamples+1):end) = 0;
dZdt(1:diffFiltDelaySamples) = [];
dZdt(1:diffFiltDelaySamples) = 0;


end