function  data = read5fs(filename, userInsertedLabelsOnly)
%% read5fs Reads in a VUDAMS 5fs file.
%
% Syntax:
%    read5fs(filename, userInsertedLabelsOnly)
%  
% Input:
%    filename --> The to be converted 5FS file.
%    userInsertedLabelsOnly --> only convert user inserted labels
%
% Output:
%    data --> Stucture that can be saved as a .physioData file
%    which can be read by the PhysioData toolbox.
%
% Iris Spruit, Elio Sjak-Shie, Leiden University, Sept 2019.
%
%--------------------------------------------------------------------------


%% Init File:

% If no filename was passed, allow the user to select one:
if nargin==0
    filename = uigetfile('.5FS');
end

% If userInsertedLabelsOnly was not specified, set it to true:
if nargin == 1 
    userInsertedLabelsOnly = true;
end


% Open the file:
fid         = fopen(filename,'r');
fileKiller  = onCleanup(@() fclose(fid));


%% Read File Header:

% Load file attributes:
s.identify                  = fread(fid,4,'*char')' ;
s.dwEndianTag               = fread(fid,1,'uint32');
s.wTag                      = fread(fid,1,'uint16');
s.wwSize                    = fread(fid,1,'uint16');
s.dwSoftwareVersion         = fread(fid,1,'uint32');
s.dwHardwareVersion         = fread(fid,1,'uint32');
s.dwSerialNumber            = fread(fid,1,'uint32');
s.szCompileDate             = fread(fid,12,'*char')';
s.szProducer                = fread(fid,64,'*char')';
s.tStamp                    = fread(fid,8,'char')';
s.dstBegin                  = fread(fid,4,'char')';
s.dstEnd                    = fread(fid,4,'char')';
s.lTimeZoneJump             = fread(fid,1,'uint32');
s.szSubjectID               = fread(fid,12,'*char')';
s.dwSession                 = fread(fid,1,'uint32');
s.dwSampleTime_us           = fread(fid,1,'uint32');
s.nChannels                 = fread(fid,1,'uint32');
s.dwFileVersion             = fread(fid,1,'uint32');
s.szReserved                = fread(fid,364,'char')';

% Load Channel attributes:
for chanIndx = 1: s.nChannels
    s.channel_info(chanIndx).wTag             = fread(fid,1,'uint16');
    s.channel_info(chanIndx).wwSize           = fread(fid,1,'uint16');
    s.channel_info(chanIndx).szID             = deblank(fread(fid,8,'*char')');
    s.channel_info(chanIndx).dwDivider        = fread(fid,1,'uint32');
    s.channel_info(chanIndx).nBits            = fread(fid,1,'uint32');
    s.channel_info(chanIndx).lMinValue        = fread(fid,1,'int32');
    s.channel_info(chanIndx).lMaxValue        = fread(fid,1,'int32');
    s.channel_info(chanIndx).lMinMaxDivider   = fread(fid,1,'int32');
    s.channel_info(chanIndx).szUnit           = deblank(fread(fid,8,'*char')');
    s.channel_info(chanIndx).dwChannelTag     = fread(fid,1,'uint16');
    s.channel_info(chanIndx).wBufSize         = fread(fid,1,'uint16');
    s.channel_info(chanIndx).bStore           = fread(fid,1,'int32');
    s.channel_info(chanIndx).szReserved       = fread(fid,16,'char');
end


%% Load Raw Data:

% Struct to hold eventdata:
eventCounter    = 1;
eventStruct     = struct('clockTick_ms',{},'type',{},'code',{},'msg',{});

offsetCounter   = 1;

generalCounter = 0;

% Read in all data:
alldata         = fread(fid,inf,'uint16=>uint16');

% Iterate through the file and extract the events and signals:
dataPointer     = 1;
channelTags     = [s.channel_info.dwChannelTag];
nChannels       = s.nChannels;
chunckTracker   = false(length(alldata),nChannels);

while dataPointer<length(alldata)
    
    generalCounter = generalCounter + 1;
    
    % Read tag:
    curTag = alldata(dataPointer);

    % Read size of incoming packet in bytes:
    curSize = double(alldata(dataPointer + 1));
    
    % Read data:
    curDataIndx = dataPointer + 1 + 1 : dataPointer + (curSize/2) - 1;
    
    % If this is an event:
    if curTag == 3
        
        % Get the cur data:
        curData = alldata(curDataIndx);
        
        % Get event type:
        eventType = typecast(curData(3:4),'uint32');

        % Process event when:
        %   * All labels need to be processed
        %   * Only user inserted labels need to be processed and the event 
        %     is user inserted (type 2) or a button press (type 0)
        if ~userInsertedLabelsOnly || eventType == 0 || eventType == 2
            
            % Get the clock time stamp:
            eventStruct(eventCounter).clockTick_ms...
                = typecast(curData(1:2),'uint32');
            
            % Get the type:
            eventStruct(eventCounter).type...
                = typecast(curData(3:4),'uint32');
            
            % Get the code:
            eventStruct(eventCounter).code...
                = typecast(curData(5:6),'int32');
            
            % Get the msg:
            fullStr = char(typecast(curData(7:22),'uint8'))';
            eventStruct(eventCounter).msg...
                = fullStr(1:find(fullStr==0,1,'first')-1);
            
            % Update counter:
            eventCounter = eventCounter + 1;
            
        end
        
    % If this is a start recording tag (is this correct??)
    elseif curTag == 4
        
        % Check if the start recording appears twice:
        if offsetCounter == 2
            error('More than 1 start of recording found')
        end
        
         % Get the cur data:
        curData = alldata(curDataIndx);   
        
        % Get the offset to the start of file
        eventsOffset = double(typecast(curData(1:2),'uint32'));  
        
        % Update counter:
        offsetCounter = offsetCounter + 1;  
        
    else
        
        % If the current chunk contains a signal, mark its location in the
        % chuck tracker logical array, under the column corresponding to
        % the current tag:
        chunckTracker(curDataIndx,curTag==channelTags) = true;
        
    end
    
    % Increment the dataPointer to collect the next block of data:
    dataPointer = dataPointer + curSize/2;
    
end




%% Process Raw Channels:

% Preallocate data spots:
channelUnits       = cell(1,nChannels);
channelNames       = cell(1,nChannels);
channelDescription = cell(1,nChannels);
channels           = cell(1,nChannels);

% Calculate the max number of samples in each channel:
maxSamples = max(sum(chunckTracker,1));

% Generate the master time vector:
fs               = 10^6/(s.dwSampleTime_us);
masterTimeVector = (0:maxSamples-1)/fs;

% Loop through the channels and process them:
for chanIndx = 1 : nChannels
    
    % Get the current data:
    curRawData = alldata(chunckTracker(:,chanIndx));
    
    % Cast data to double:
    curData = double(curRawData);
    
    % SCL channel uses different scaling:
    if strcmp(s.channel_info(chanIndx).szID, 'SCL')
        
        
        % Non-Linear GSR mapping table:
        nbits = s.channel_info(chanIndx).nBits;
        dcList = [...
            0    , 0     ;...
            100  , 771   ;...
            500  , 3803  ;...
            1000 , 7449  ;...
            1484 , 10858 ;...
            2000 , 14408 ; ...
            3030 , 20962 ;...
            5000 , 32487 ;...
            6993 , 42516 ;...
            10000, 55631 ];
        lowerBound = -(2^(nbits - 1));
        upperBound = 2^(nbits - 1) - 1;
        convList2 = interp1(...
            dcList(:,2) ...
            ,dcList(:,1)...
            ,(0:(upperBound - lowerBound))'...
            ,'linear',dcList(end,1))/100;
        curData = convList2(curData+1);
        
    else
        
        % Scale the data:
        minVal   = s.channel_info(chanIndx).lMinValue;
        maxVal   = s.channel_info(chanIndx).lMaxValue;
        valRange = maxVal - minVal;
        rawMax   = 2^s.channel_info(chanIndx).nBits - 1;
        curData  = minVal + ( valRange .* (curData ./ rawMax) );
        
    end
    
    % Calculate channelFs, which reflects the actual sampling rate of the
    % current channel
    curFs = fs/s.channel_info(chanIndx).dwDivider;
    
    % Check if interpolation is required:
    % It is assumed that curFs is always smaller than fs and also a
    % whole fraction of fs. As such, we introduce half dt delay when
    % into the current time vector when upsampling the current signal to
    % the master time vector.
    if curFs ~= fs
        
        % Interpolate the data to the master time vector (perform linear
        % intrapolation, with nearest neighbor extrapolation):
        curTimeVector = (0.5 * 1/curFs) + (0:length(curData)-1)/curFs ;
        curSig = interp1(curTimeVector, ...
            curData,masterTimeVector,'linear');
        curSigExtrap = interp1(curTimeVector, ...
            curData,masterTimeVector,'nearest','extrap');
        curSig(isnan(curSig)) = curSigExtrap(isnan(curSig));
        channels{chanIndx} = curSig';
        
    else
        
        % If no interpolation is required, write the data straight to the
        % channels cell:
        channels{chanIndx} = curData;
        
    end
    
    % Extract metadata from struct array:
    channelNames{chanIndx} ...
        = strrep(s.channel_info(chanIndx).szID...
        , char(0) ,'');
    channelUnits{chanIndx} ...
        = strrep(s.channel_info(chanIndx).szUnit...
        , char(0) ,'');
    
    
    % Correct the ECG case:
    if strcmp(channelNames{chanIndx},'ECG')
        channels{chanIndx} = channels{chanIndx}./3.6;
    end
    
    % Generate description:
    channelDescription{chanIndx} = ...
        sprintf('From VUDAMS channel %i (fs=%i)',chanIndx,curFs);
    
end


%% Generate the absolute acceleration vector:
% Only convert the movement channels when all movement channels exist

Xind = strcmp(channelNames,'MXR');
Yind = strcmp(channelNames,'MYR');
Zind = strcmp(channelNames,'MZR');

if any(Xind) && any(Yind) && any(Zind)
    
    % Calculate absolute accel:
    XYZ = sqrt(channels{Xind}.^2 + channels{Xind}.^2 + channels{Xind}.^2);
   
    % Add channel:
    channels{end+1}            = XYZ;
    channelDescription{end+1}  = 'Absolute Accel. XYZ';
    channelNames{end+1}        = 'XYZ';
    channelUnits{end+1}        = 'g';
    
    % Delete the old channels:
    accelChannels =  [Xind | Yind | Zind false];
    channels(accelChannels)            = [];
    channelDescription(accelChannels)  = [];
    channelNames(accelChannels)        = [];
    channelUnits(accelChannels)        = [];
    
end


%% Generate the dZ/dt channel:

% Check if the Z0 channel is present:
if ~isempty(strcmp(channelNames, 'Z0'))
    
    % Settings:
    diffFIR_order = 256;
    diffFIR_HPCF  = 1;
    diffFIR_LPCF  = 10;
    
    % Get Z0:
    Z0 = channels{strcmp(channelNames,'Z0')};
    
    % Calculate the dZ/dt signal from the Z0 signal
    DZDT = getDZDT(Z0, diffFIR_order, diffFIR_HPCF, diffFIR_LPCF, fs);
    
    % Add channel:
    channels{end+1}           = DZDT;
    channelUnits{end+1}       = 'Ohm/sec';
    channelNames{end+1}       = 'DZDT';
    channelDescription{end+1} = 'Calculated from Z0';
    
end


%% Save all data:

% Insert signal data into master struct:
data.signals.fs                 = fs;
data.signals.channels           = channels;
data.signals.channelUnits       = channelUnits;
data.signals.channelNames       = channelNames;
data.signals.channelDescription = channelDescription;

% Fetch the offset from the dump (uint32 at location 875 to 876):
% eventsOffset = double(typecast(s.dump(875:876),'uint32'));

% Generate the event data (as labels):
data.labels.t                   = ...
    ( double( vertcat( eventStruct.clockTick_ms ) ) ...
    - eventsOffset ) / 1000;
data.labels.channels{1}            = ...
    strcat( {'Label (Type: '}...
    , num2str(vertcat(eventStruct.type))...
    ,{', Code: '}...
    ,strtrim(num2str(vertcat(eventStruct.code)))...
    ,{', Msg: '}...
    ,{eventStruct.msg}');

data.labels.channelUnits{1}   = '-';
data.labels.channelNames{1}   = 'Labels';

