function [data] = readPhysLog(logFilename,doECG,doPPU,doRESP,doGRAD)
%convertScannerLog Converts physio-log files into physioData files.
%
%   data = readPhysLog(logFilename[,doECG,doPPU,doRESP,doGRAD])
%
%    The function reads a log file (logFilename) and converts it to a
%    physioData file-compatible data struct. The v1, v2, ppu and resp
%    timeseries are converted to channels.
%
%    See the included LICENSE.txt file.
%
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Changelog:
%
%   Oct-2016 -- Elio:
%        > Wrote first version.
%
%   Oct-2016 -- Elio & Iris Spruit:
%        > Updated for the new LIBC Scanner data format. Not backwards
%          compatible.
%
%   Mar-2018 -- Elio:
%        > Rewrote the script, now compatible with both LIBC scanner
%          versions. Markers are now extracted using proper hex to dec to
%          bit conversion. The physiological trigger markers are now
%          transformed into vcg, ppu and resp channels, with pulses placed
%          at the triggers times.
%
%--------------------------------------------------------------------------


%% Define Paramerts and Parse Inputs:

% Scanner sampling freq:
fs = 500;

% Flags to activate extraction of ECG, PPU, RESP and the gradient channels:
if nargin < 2
    doECG   = true;
    doPPU   = true;
    doRESP  = true;
    doGRAD  = true;
end

% Init return struct:
data = struct();


%% Read log File

% Open file:
fid       = fopen(logFilename);
if fid < 1
    error('Could not open file!');
end
killFile  = onCleanup(@() fclose(fid));

% Get data (skip lines that begin with #'s, and everything after the 10th
% column):
rawData = textscan(fid,'%f %f %f %f %f %f %f %f %f %s %*[^\n]'...
    ,'CommentStyle','#');

% Extract data:
mark       = rawData{10};
nSamples   = numel(mark);


%% Process Markers:

%--------------------------------------------------------------------------
%
% The 'mark' channel bits are defined as (note that the bits here are
% zero-indexed, whereas matlab uses one-based indexes):
%
% * Bit  0: 0001 - VAPHYS_MARKER_TRIGGER_ECG
%                  Actual ECG trigger used by measurement.
%
% * Bit  1: 0002 - VAPHYS_MARKER_TRIGGER_PPU
%                  Actual PPU trigger used by measurement.
%
% * Bit  2: 0004 - VAPHYS_MARKER_TRIGGER_RESP
%                  Actual RESP trigger used by measurement
%
% * Bit  3: 0008 - VAPHYS_MARKER_MEASUREMENT
%
% * Bit  4: 0010 - VAPHYS_MARKER_SCAN_BEGIN
%                  Start scan.

% * Bit  5: 0020 - VAPHYS_MARKER_SCAN_END
%                  End scan.
%
% * Bit  6: 0040 - VAPHYS_MARKER_TRIGGER_EXTERNAL
%                  External trigger.
%
% * Bit  7: 0080 - VAPHYS_MARKER_CAL_AUTO
%                  Timepoint for automatic calibration.
%
% * Bit  8: 0100 - VAPHYS_MARKER_MANUAL_START
%                  Manual start indication.
%
% * Bit  9: 0200 - VAPHYS_MARKER_CAL_MANUAL_OUT
%                  Timepoint Outside Manual calibration.
%
% * Bit 10: 0400 - VAPHYS_MARKER_CAL_MANUAL_IN
%                  Timepoint Inside Manual calibration.
%
% * Bits 11 - 15 not used by this script.
%
% NOTE: Timepoints calibration means that data used for calibration is
% before this timepoint, so it is infact pointing to the last sample of the
% data used for calibration.
%
%--------------------------------------------------------------------------

% Convert markers into 16-bit unsigned integer, and generate a time vector:
markDec = uint16(hex2dec(mark));
t       = timeVector(fs,nSamples);

% Transform the physiological markers into channels with spikes where the
% markers where send:
if doECG || doPPU || doRESP
    triggerSlopeTime   = 0.1;
    convKernel         = [linspace(0,1,triggerSlopeTime * fs) ...
        1.1 linspace(1,0,triggerSlopeTime * fs)];
end
if doECG
    ECG_TRIG           = conv(double(bitget(markDec,1)),convKernel,'same');
end
if doPPU
    PPU_TRIG           = conv(double(bitget(markDec,2)),convKernel,'same');
end
if doRESP
    RESP_TRIG          = conv(double(bitget(markDec,3)),convKernel,'same');
end

% Extract the timestamps of the abovementioned bits, for the events that
% are to be turned into labels:
markers.MEASUREMENT     = t(logical(bitget(markDec,4)));
markers.SCAN_BEGIN      = t(logical(bitget(markDec,5)));
markers.SCAN_END        = t(logical(bitget(markDec,6)));
markers.TRIGGER_EXT     = t(logical(bitget(markDec,7)));
markers.CAL_AUTO        = t(logical(bitget(markDec,8)));
markers.MANUAL_START    = t(logical(bitget(markDec,9)));
markers.CAL_MANUAL_OUT  = t(logical(bitget(markDec,10)));
markers.CAL_MANUAL_IN   = t(logical(bitget(markDec,11)));

% Make labels from the markers struct by collecting their fieldnames and
% timestamps:
markerLabels = fieldnames(markers);
[markerTimes, markerLabels] = cellfun(@(f) ...
    deal({markers.(f)},{repmat({f},length(markers.(f)),1)}),markerLabels);

% Make labels:
if ~all(cellfun(@isempty,markerTimes))
    
    % Get all timestamps, and sort them:
    data.labels.t               = vertcat(markerTimes{:});
    [data.labels.t,sortIndx]    = sort(data.labels.t);
    
    % Get all labels, and order them with the sort from above:
    data.labels.channels{1}     = vertcat(markerLabels{:});
    data.labels.channels{1}     = data.labels.channels{1}(sortIndx);
    
    % Add necessary label metadata:
    data.labels.channelUnits{1} = '-';
    data.labels.channelNames{1} = 'Labels';
    
end


%% Make Channels:

% Option to multiply the physio channels with a gain to scale them:
preGain = 1;

% Offset the fake trigged-derived physio pulse channels with a small
% constant value to prevent the Toolbox from assuming its a marker channel
% (when its empty):
trigChanOffset = -0.1;

% ECG:
chanN = 0;
if doECG
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'V1';
    data.signals.channels{chanN}            = preGain .* rawData{3};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'V1 from scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'V2';
    data.signals.channels{chanN}            =  preGain .* rawData{4};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'V2 from scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'V1+V2';
    data.signals.channels{chanN} ...
        =  preGain .* (rawData{3} + rawData{4});
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'V1 + V2 from scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'ECG_TRIG';
    data.signals.channels{chanN}            =  ECG_TRIG - trigChanOffset;
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'ECG triggers from scanner';
end

% PPU:
if doPPU
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'PPU';
    data.signals.channels{chanN}            = preGain .* rawData{5};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'PPU From Scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'PPU_TRIG';
    data.signals.channels{chanN}            = PPU_TRIG - trigChanOffset;
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'PPU triggers from scanner';
end

% RESP:
if doRESP
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'RESP';
    data.signals.channels{chanN}            = rawData{6};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'RESP from scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'RESP_TRIG';
    data.signals.channels{chanN}            = RESP_TRIG - trigChanOffset;
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'RESP triggers from scanner';
end

% Gradients:
if doGRAD
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'GRADIENT_X';
    data.signals.channels{chanN}            = rawData{7};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'Gradient X from scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'GRADIENT_Y';
    data.signals.channels{chanN}            = rawData{8};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'Gradient Y from scanner';
    
    chanN                                   = chanN + 1;
    data.signals.channelNames{chanN}        = 'GRADIENT_Z';
    data.signals.channels{chanN}            = rawData{9};
    data.signals.channelUnits{chanN}        = '-';
    data.signals.channelDescription{chanN}  = 'Gradient z from scanner';
end

% Save sampling frequency:
if doECG || doPPU || doRESP || doGRAD
    data.signals.fs = fs;
end


end






