function [dataTable, diamData, info] = read_EET(fn,doOnlyDiam,numNames)
%read_EET Reads an E-Prime EET output file.
%
%  This function reads one EET tab-seperated text-file, determines the
%  version, and outputs the data in a table and a seperate struct
%  containing only the pupil diameter data (formatted for use with the
%  PhysioData Toolbox).
%
%--------------------------------------------------------------------------
%  Calling convention:
%
%    [dataTable, diamData, info] = read_EET(fn[,doConvert][,numNames])
%
%--------------------------------------------------------------------------
%  Input:
%
%    fn         : The full filename to the EET .gazedata or .txt file.
%
%    doConvert  : (OPTIONAL) Default is true. If set to true, this function
%                 will search for standard numeric columns (e.g. time, size
%                 and coordinate data), and convert them to doubles. Set to
%                 false to speed up the function when the other data will
%                 not be used anyways.
%
%    NumNames   : OPTIONAL: a 1 x n cell array of strings specifying the
%                 names of additional data columns to convert from strings
%                 to doubles. The names of the columns are taken from the
%                 first row of the data file. The columns specified here
%                 will always be converted, regardless of the coConvert
%                 flag.
%
%--------------------------------------------------------------------------
%  Output:
%
%    dataTable  : A table containing the EET data. Note that data rows are
%                 sorted by timestamps, and rows with duplicate timestamps
%                 are removed by deleting the preceding duplicates (i.e.,
%                 the last unique timestamp is kept). This makes sure that
%                 the data timevector is strictly monotonically increasing.
%
%                 Additionally, a column named 't_ms' is added, which is
%                 the zeroed timevector that should be used as the
%                 timestamps for the rows, and which is in the same
%                 timevector as the data in diamData. The zeroed timevector
%                 is generated by subtracing the value of the first
%                 timestamp, which has an arbitrary value, from the rest.
%
%                 Variables that are considered to be numeric, such as
%                 times, distances, coordinates, etc., are transformed into
%                 numeric data, all else is left as strings.
%
%    diamData   : A struct containing diameter data, wiht the timestamps in
%                 the filed t_ms, and the left and right pupil diameters,
%                 in mm, in the fields L and R, respectively. Rows where no
%                 pupil data is available are not included. When data of a
%                 single pupil is missing or invalid, that pupil's size is
%                 set to NaN.
%
%    info       : A struct containing info about the data.
%
%--------------------------------------------------------------------------
%  NOTES:
%
%    The files must use dot decimal seperators. Comma seperators are not
%    supported and will result is faulty data (possibly w/o this function
%    throwing an error).
%
%    This function has been tested on Windows 10 using the following
%    E-Prime, EET, and eye-tracker combinations: 
%     # Tobii X2-60   &  E-Prime 2.0.10.356  &  EET 2.0.2.54
%     # Tobii X2-60   &  E-Prime 3.0.3.60    &  EET 3.1.1.56
%     # Tobii X2-60   &  E-Prime 3.0.3.80    &  EET 3.2.1.105
%     # Tobii X2-120  &  E-Prime 2.0.10.356  &  EET 2.0.2.54
%     # Tobii X2-120  &  E-Prime 3.0.3.60    &  EET 3.1.1.56
%     # Tobii X2-120  &  E-Prime 3.0.3.80    &  EET 3.2.1.105
%
%--------------------------------------------------------------------------
%   Part of the PhysioData Toolbox.
%    https://physiodatatoolbox.leidenuniv.nl/
%     Elio Sjak-Shie, July 2019.
%
%  Thanks to Iris Spruit for providing EET documentaion and test data.
%--------------------------------------------------------------------------


%% Open File and Read Header:

% Set doConvert
if nargin < 2
    doOnlyDiam = true;
end

% Set numNames:
if nargin < 3
    numNames = {};
else
    assert(iscell(numNames) && size(numNames,1) == 1 ...
        ,'numNames must be a 1 x n array of strings.')
end

% Open file:
fileID = fopen(fn,'r');
if fileID < 1
    error('Could not open file.')
end
fileCloser = onCleanup(@() fclose(fileID));

% Read header line from file:
headerCell = textscan(fileID,'%[^\n\r]',1);
headers    = strsplit(headerCell{1}{1},'\t');


%% Determine Gazedata Version

% Make a struct defining the versions and the name of the time, diameter
% and validity columns (note that if the time is not in ms, it should be
% corrected later on):
verInfo = struct(...
    'v20'...
    ,struct(...
    'time','TETTime'...
    ,'L','DiameterPupilLeftEye'...
    ,'R','DiameterPupilRightEye'...
    ,'valL','ValidityLeftEye'...
    ,'valR','ValidityRightEye'...
    )...
    ...
    ,'v31'...
    ,struct(...
    'time','HardwareTimestamp'...
    ,'L','GazeOriginDiameterPupilLeftEye'...
    ,'R','GazeOriginDiameterPupilRightEye'...
    ,'valL','GazeOriginValidityPupilLeftEye'...
    ,'valR','GazeOriginValidityPupilRightEye'...
    )...
    ...
    ,'v32'...
    ,struct(...
    'time','HardwareTimestamp'...
    ,'L','PupilDiameterLeftEye'...
    ,'R','PupilDiameterRightEye'...
    ,'valL','PupilValidityLeftEye'...
    ,'valR','PupilValidityRightEye'...
    )...
    );

% Function to search for the variable names in the headers:
findVars = @(f) structfun(@(h) strcmp(headers,h),f,'UniformOutput',false);

% Function to test if the variable names have been found:
areVarsThere = @(f) structfun(@(v) sum(v)==1,f,'UniformOutput',true);

% Loop thru the versions and check the column names defined above are
% present:
verNames =  fieldnames(verInfo)';
eetVer   = '';
for curVerNameC = verNames
    curVerName = curVerNameC{1};
    varData = findVars(verInfo.(curVerName));
    if all(areVarsThere(varData))
        eetVer = curVerName;
        break
    end
end
if isempty(eetVer)
    error('The gazedata file had an unknown format.');
end

% Clear vars that should not be used anymore:
clearvars findVars areVarsThere verNames headerCell


%% Read and Preprocess Data:

% If only the diameter data is to be extracted, disable all channels by
% default; otherwise if all data is to be extracted, enable all channels as strings by
% default, then set the known numeric columns to numeric:
nCols = numel(headers);
if doOnlyDiam
    txtScanFormC = repmat({'%*s'},1,nCols);
    colsUsed     = false(nCols,1);
else
    txtScanFormC = repmat({'%s'},1,nCols);
    txtScanFormC(isNumeric(headers,numNames)) = {'%f'};
    colsUsed     = true(nCols,1);
end

% Set the version's diam, time, and validity columns to numeric:
diamCols = structfun(@(v) find(v),varData,'UniformOutput',true);
txtScanFormC(diamCols) = {'%f'};
colsUsed(diamCols)     = true;

% Read the data. Note that newer EET versions have -1.#IND or -1.#IND0 in
% the data to indicate NaNs, as such, specify that it should be treated as
% missing data:
rawDataArray = textscan(fileID,[txtScanFormC{:}]...
    ,'Delimiter', '\t'...
    ,'TreatAsEmpty',{'-1.#IND0' '-1.#IND'}...
    ,'EmptyValue',NaN);
cellNumels = cellfun(@(c) {numel(c)}, rawDataArray);
assert(~isequal(0,cellNumels{:}),'No data.');
assert(isequal(cellNumels{:}),'Unequal number of data rows.');

% Check if the end of the file was reached:
if ~feof(fileID)
    error('File was not read correctly.');
end

% Fix the headers to that they are valid variable names for the table:
varNameHeaders = matlab.lang.makeUniqueStrings(...
    matlab.lang.makeValidName(headers(colsUsed)));

% Make the rawData table:
rawDataTable = table(rawDataArray{:}...
    ,'VariableNames',varNameHeaders);

% Clear vars that should not be used anymore:
clearvars headers rawDataArray varNameHeaders nCols txtScanFormC

% Depending on the version, perform the correct action in order to produce
% a table with the raw data (rawDataTable):
switch eetVer
    case 'v20'
        % If this looks like a v2.0 file:
        
        % Add a variable that represents the time in ms:
        rawDataTable.tRaw_ms = rawDataTable.(verInfo.(eetVer).time);
        
    case {'v31' 'v32'}
        % If this looks like a v3.x file:
        
        % Add a variable that represents the time in ms (note, in these
        % versions it is originally in us):
        rawDataTable.tRaw_ms ...
            = rawDataTable.(verInfo.(eetVer).time) ./ 1000;
        
    otherwise
        error('Unknown gazedata version.');
end

% Close the file:
delete(fileCloser);

% Sort time and remove time duplicates from table (note, preceding repeats
% are marked as dupes and removed; i.e., the last occurence of each
% timestamps is kept):
[t_ms_sort,sortedTimeIndx] = sort(rawDataTable.tRaw_ms);
timeDupes                  = [~(diff(t_ms_sort) > 0); false];
dataTable                  = rawDataTable(sortedTimeIndx(~timeDupes),:);

% Determine the zero time, and add a zeroed time vector:
zeroTime_ms    = dataTable.tRaw_ms(1);
dataTable.t_ms = dataTable.tRaw_ms - zeroTime_ms;

% Count the number of rows in the original data:
rawDataRowsCount = height(rawDataTable);

% Clear data that should no longer be used:
clearvars rawDataTable t_ms_sort sortedTimeIndx timeDupes rawDataArray ...
    varData fileCloser fileID findVars readData areVarsThere


%% Process Diameters:

% Get the diameter data from the variablenames determined previously for
% this verson, except the time, which should now be taken from the the new
% zeroed time field (t_ms):
varTableData      = rmfield(verInfo.(eetVer),'time');
varTableData.t_ms = 't_ms';
diamData = structfun(@(f) dataTable.(f),varTableData...
    ,'UniformOutput',false);

% Depending on the version, fix the diameter:
switch eetVer
    case 'v20'
        % This version of EET uses the following validity ratings:
        %
        %    0: The system is certain that it has recorded all
        %       relevant data for the particular eye.
        %
        %    1: The system has only recorded one eye and has made some
        %       assumptions and estimations regarding if the recorded eye
        %       is left or right. However, it is still highly probable that
        %       the estimations done are correct.
        %
        %     2: The system has only recorded one eye and has no way of
        %        determining if this is the left or the right eye.
        %
        %     3: The system is fairly confident that the actual gaze data
        %        is incorrect or corrupted.
        %
        %     4: The actual gaze data is missing or incorrect.
        %
        % As such, convert only values 0 an 1 to valid, and the rest to
        % invalid:
        diamData.valL = ismember(diamData.valL,[0 1]);
        diamData.valR = ismember(diamData.valR,[0 1]);
        
    case {'v31' 'v32'}
        % Nothing special necessary vor v3x. Note that the time, originally
        % in us, has already been corrected to ms:
    otherwise
        error('Unknown gazedata version.');
end

% Make function to test validity (Note, some raw diameter data may have
% been saved as empty, -1.#IND, etc., but these could not be transformed
% into numbers, and so should already be NaNs):
isInvalid = @(f) diamData.(f) < 0 | ~diamData.(['val' f]);

% Set invalid samples to NaN:
diamData.L(isInvalid('L')) = NaN;
diamData.R(isInvalid('R')) = NaN;

% Detect rows with no diameter data (left and right are invalid), and
% delete them:
noDiameterDataRows = all(isnan(diamData.L) & isnan(diamData.R), 2);
diamData = structfun(@(f) f(~noDiameterDataRows)...
    ,diamData,'UniformOutput',false);

% Remove the validity rows, as they are now represneted by NaNs in the
% diameter data:
diamData = rmfield(diamData,{'valL' 'valR'});


%% Add Info:

% Create a struct with info about the file:
info.eetVer           = eetVer;
info.varMap           = verInfo.(eetVer);
info.zeroTime_ms      = zeroTime_ms;
info.max_t_ms         = dataTable.t_ms(end);
info.rawDataRowsCount = rawDataRowsCount;


end


function numCols = isNumeric(headers,numNames)
% Finds the columns that are numeric.

numericDataColumnNames = [numNames ...
    'ACC'...
    ,'CurrentFixationDuration'...
    ,'CursorX'...
    ,'CursorY'...
    ,'DistanceLeftEye'...
    ,'DistanceRightEye'...
    ,'GazeOriginDiameterPupilLeftEye'...
    ,'GazeOriginDiameterPupilRightEye'...
    ,'GazeOriginPositionTrackBoxXLeftEye'...
    ,'GazeOriginPositionTrackBoxXRightEye'...
    ,'GazeOriginPositionTrackBoxYLeftEye'...
    ,'GazeOriginPositionTrackBoxYRightEye'...
    ,'GazeOriginPositionTrackBoxZLeftEye'...
    ,'GazeOriginPositionTrackBoxZRightEye'...
    ,'GazeOriginPositionUserXLeftEye'...
    ,'GazeOriginPositionUserXRightEye'...
    ,'GazeOriginPositionUserYLeftEye'...
    ,'GazeOriginPositionUserYRightEye'...
    ,'GazeOriginPositionUserZLeftEye'...
    ,'GazeOriginPositionUserZRightEye'...
    ,'GazeOriginValidityLeftEye'...
    ,'GazeOriginValidityRightEye'...
    ,'GazePointPositionDisplayXLeftEye'...
    ,'GazePointPositionDisplayXRightEye'...
    ,'GazePointPositionDisplayYLeftEye'...
    ,'GazePointPositionDisplayYRightEye'...
    ,'GazePointPositionUserXLeftEye'...
    ,'GazePointPositionUserXRightEye'...
    ,'GazePointPositionUserYLeftEye'...
    ,'GazePointPositionUserYRightEye'...
    ,'GazePointPositionUserZLeftEye'...
    ,'GazePointPositionUserZRightEye'...
    ,'GazePointValidityLeftEye'...
    ,'GazePointValidityRightEye'...
    ,'HostTimestamp'...
    ,'ID'...
    ,'LineNumber'...
    ,'PupilDiameterLeftEye'...
    ,'PupilDiameterRightEye'...
    ,'RT'...
    ,'RTTime'...
    ,'RTTimeMicro'...
    ,'RTTimeReceived'...
    ,'RTTimeReceivedMicro'...
    ,'RunningSample'...
    ,'SEQ'...
    ,'SEQD'...
    ,'SEQH'...
    ,'Session'...
    ,'Subject'...
    ,'SystemTimestamp'...
    ,'TimestampMicrosec'...
    ,'TimestampSec'...
    ,'XCameraPosLeftEye'...
    ,'XCameraPosRightEye'...
    ,'XGazePosLeftEye'...
    ,'XGazePosRightEye'...
    ,'YCameraPosLeftEye'...
    ,'YCameraPosRightEye'...
    ,'YGazePosLeftEye'...
    ,'YGazePosRightEye'];

numCols = ismember(headers,numericDataColumnNames);

end


