function [bdfHeader,signals] = read_bdf(filename,doData)
% Reads a BIOSEMI bdf file, and returns the header and a signals struct for
% use in the .hysioData Toolbox
%
%  Call:
%    [bdfHeader,signals] = read_bdf(filename[,doData]);
%
%    where doData is an optional flag to enable the reading of the actual
%    data. The function will only return metadata if set to false. Default
%    is true.
%
% See: https://www.biosemi.com/faq/file_format.htm
%
% Elio Sjak-Shie, July 2019.
%--------------------------------------------------------------------------


%% Read File:

% Process inputs:
if nargin < 2
    doData = true;
end

% Open file:
fid = fopen(filename);
killFile = onCleanup(@() fclose(fid));
if fid == -1
    error('Could not open file.');
end

% Read header:
readChar = @(len) deblank(fread(fid,len,'char=>char')');
fseek(fid,8,'bof');
bdfHeader.subjectNum = readChar(80);
bdfHeader.recNum     = readChar(80);
bdfHeader.startDate  = readChar(8);
bdfHeader.startTime  = readChar(8);

% Rec num:
fseek(fid, 236, 'bof');
readNum = @(len) str2double(deblank(fread(fid,len,'char=>char')'));
numRec  = readNum(8);
if numRec == -1
    error('Unknown record number not supported.');
end

% Rec duration and channel info:
recDur                        = readNum(8);
nChans                        = readNum(4);
assert(recDur==1 ...
    ,'Only a recording duration of 1 second is currently supported.');

% Get channel data:
readStrFromChans = @(len) cellstr(fread(fid,[len nChans],'char=>char')')';
signals.channelNames            = readStrFromChans(16);
signals.channelTransducerTypes  = readStrFromChans(80);
signals.channelUnits            = readStrFromChans(8);
signals.channelPhysicalMin      = str2double(readStrFromChans(8));
signals.channelPhysicalMax      = str2double(readStrFromChans(8));
signals.channelDigitalMin       = str2double(readStrFromChans(8));
signals.channelDigitalMax       = str2double(readStrFromChans(8));
signals.channelPrefiltering     = readStrFromChans(80);
signals.channelDescription      = strcat(signals.channelNames...
    ,{': '},signals.channelPrefiltering);

% Get the samples per record:
samplesPerRecordAll                = readStrFromChans(8);
assert(isequal(samplesPerRecordAll{:})...
    ,'Only channels with one sampling rate is currently supported.');
signals.fs = str2double(samplesPerRecordAll{1});
signals.L                       = recDur * numRec * signals.fs;


%% Do Data if Requested:

if doData
    
    % Skip the reserved section:
    fseek(fid,32*nChans,0);
    
    % Preallocate the data and fetch it from file. Note that in this
    % version of the code, recDur must be one, so fs is the number of
    % samples in that rec.
    numSam  = signals.fs;
    data    = NaN(numRec * recDur * numSam, nChans);
    for ind = 1:numRec
        auxL = [recDur * numSam,nChans];
        data((((ind-1)*auxL)+1):(ind*auxL),:) = fread(fid,auxL,'bit24');
    end
    
    % Scale the data
    channelGain = ...
        (signals.channelPhysicalMax...
        - signals.channelPhysicalMin)...
        ./...
        (signals.channelDigitalMax...
        - signals.channelDigitalMin);
    signals.channels = num2cell((data .* channelGain),1);
    
end

delete(killFile);

end
