function [structOut] = importPDTxlsx(filename)
%pdtxls2table Imports the fancy PhysioDataToolbox Excel file as a table.
%
% Use this function to read the Excel-table produced by the Toolbox into
% MATLAB. Alternatively, export MATLAB tables directly from the Toolbox.
%
% Elio Sjak-Shie
%--------------------------------------------------------------------------

% Premake output struct:
structOut = struct();

% Present file selection gui if necessary:
if nargin < 1
    [f,p] = uigetfile('*.xlsx');
    if f == 0 
        return
    end
    filename = [p f];
end

% Read file info:
fprintf('\nProcessing file: %s.\n',filename);
[~,sheeeet,~] = xlsfinfo(filename) ;
for curSheetName = sheeeet
    try
       curTable = readSheet(filename,curSheetName{1});
       if ~isempty(curTable)
           structOut.(curSheetName{1}) = curTable;
           fprintf(' > Imported sheet: %s.\n',curSheetName{1});
       end
    catch
        fprintf(2,' > Failed to import sheet: %s.\n',curSheetName{1});
    end    
end
fprintf('Done.\n\n.');
end


%% Helper Functions:

%==========================================================================
function tableOut = readSheet(filename,sheetname)

% Excel Format Parameters:
varNamesRowNum       = 5;
unitsNamesRowNum     = 3;
firstDataRow         = 6;
varDescFromBottom    = 2;
lastDataRowFromEnd   = 6;

% Read data:
[~,~,xlsRaw]  = xlsread(filename,sheetname);

% Check invalidity (sheet is empty or the first is not the Toolbox tag):
if isequal(xlsRaw,{NaN}) || ~strcmp(xlsRaw{1},'PhysioData Toolbox')
    tableOut = [];
    return
end

% Extract info:
varUnits  = xlsRaw(unitsNamesRowNum,2:end);
varDesc   = xlsRaw(end-varDescFromBottom,2:end);
rowNames  = xlsRaw((firstDataRow):end-lastDataRowFromEnd,1);
varNames  = xlsRaw(varNamesRowNum,2:end);
fileInfo  = horzcat(xlsRaw{1},' ', xlsRaw{2},'. '...
    , xlsRaw{3},' ',xlsRaw{4});

% Strip header and footer:
tableOut = cell2table(xlsRaw((firstDataRow:end-lastDataRowFromEnd),2:end)...
    ,'Variablenames',varNames,'RowNames',rowNames);

% Replace {NaN} with {''};
for curVar = tableOut.Properties.VariableNames
    if iscell(tableOut.(curVar{1}))
        tableOut.(curVar{1})(cellfun(@(c) isnumeric(c)&&isnan(c)...
            ,tableOut.(curVar{1})))   = {''};
    end
end

% Fix missing values:
varDesc(cellfun(@(c) isnumeric(c)&&isnan(c),varDesc))   = {''};
varUnits(cellfun(@(c) isnumeric(c)&&isnan(c),varUnits)) = {''};

% Save metadata:
tableOut.Properties.Description          = fileInfo;
tableOut.Properties.VariableDescriptions = varDesc;
tableOut.Properties.VariableUnits        = varUnits;

end


