function tableOut = tableVertCat(varargin)
% tableVertCat vertically concatenates tables with different variables.
%
%   tableOut = tableVertCat(Table_1,Table_2, Table_3, etc.) vertically
%   concatenates the tables passed to the function as arguments. All tables
%   must have single column variables of the following datatypes: cell,
%   double and categorical. If multiple tables have the same
%   variable name, those variables must be of the same type.
%
%--------------------------------------------------------------------------
%   Part of the PhysioData Toolbox.
%    Elio Sjak-Shie, Faculty of Social Sciences,
%     Leiden University, 2017.
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Changelog:
%
%   June-2017 -- Elio:
%        > Wrote first version.
%
%   Oct-2017 -- Elio:
%        > Added empty argument discarding functionality.
%
%   Feb-2018 -- Elio:
%        > Added unit and description copying functionality.
%
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   To do:
%
%   > Improve documentaion.
%
%   > Make function accept variables with more than 1 column.
%
%   > Replace struct for missing with map (?).
%
%   > Add support for different datatypes (?).
%
%--------------------------------------------------------------------------

% Check inputs:
assert(all(cellfun(@(T) isa(T,'table')||isempty(T),varargin))...
    ,'Input arguments must be tables, or empty')

% Delete empty arguments:
tablesIn = varargin(~cellfun(@isempty,varargin));

% If only a single valid table was passed, return it:
if length(tablesIn) == 1
    tableOut = tablesIn{1};
    return
end

% If no valid tables were passed, return an empty table:
if isempty(tablesIn)
    tableOut = table();
    return
end

% Check datatypes (doubles are forced since concatenating tables with
% different numeric types produces unexpected results due to casting):
supportedDataTypes = {'categorical' 'cell' 'double'};
assert(all(cellfun(@(T) all(varfun(@(var) ismember(class(var)...
    ,supportedDataTypes),T,'OutputFormat','uniform')),tablesIn))...
    ,'Tables may only contain categorical, cell and double data types.');

% Try normal concatenation if the tables agree:
varNames = cellfun(@(T) {sort(T.Properties.VariableNames)},tablesIn);
if isequal(varNames{:})
    tableOut = vertcat(tablesIn{:});
else
    % Otherwise, check that all variables are vectors (i.e. the second
    % dimension is a singleton):
    assert(all(cellfun(@(t) ...
        all(varfun(@(x) size(x,2)==1,t,'OutputFormat','uniform'))...
        ,tablesIn))...
    ,'All variables must have a single column.');
    
    % Specify which values to use when expanding variables:
    emptyTypeValue.double      = NaN;
    emptyTypeValue.categorical = categorical({''});
    emptyTypeValue.cell        = {''};
        
    % Map the variable names to the variable types:
    nameTypeMapCell = cellfun(@(t) ...
        containers.Map(t.Properties.VariableNames...
        ,varfun(@class,t,'OutputFormat','cell')...
        , 'uniformValues',true)...
        ,tablesIn,'UniformOutput',false)';
    nameTypeMap = vertcat(nameTypeMapCell{:});
    
    % Loop through tables and expand them so that they include all variable
    % names:
    expandedTables = cell(nargin,1);
    for curTableIndx = 1:numel(tablesIn)
        
        % Fetch current table, and determine missing variables:
        curTable = tablesIn{curTableIndx};
        curMissingVars = setdiff(nameTypeMap.keys...
            ,curTable.Properties.VariableNames);

        % Generate missing variables, and append table:
        emptyVars = cellfun(...
            @(f) repmat(emptyTypeValue.(f),size(curTable,1),1)...
            ,values(nameTypeMap,curMissingVars),'UniformOutput',false);
        expandedTables{curTableIndx} = [curTable ...
            table(emptyVars{:},'VariableNames',curMissingVars)];
    end
    
    % Concatenate tables:
    tableOut = vertcat(expandedTables{:});
    
end

% When concatenating tables, matlab seems to automatically use the
% variables descriptions and units from the first table. As such, descs and
% units may get lost in not present in the first table. The code below
% replaces emty units and descs with the first available one:
if ~isempty(tableOut.Properties.VariableDescriptions)
    tableOut = setMissingMetadata(tableOut,tablesIn...
    ,'VariableDescriptions');
end

if ~isempty(tableOut.Properties.VariableUnits)
    tableOut = setMissingMetadata(tableOut,tablesIn...
    ,'VariableUnits');
end

end

%==========================================================================
function tableOut = setMissingMetadata(tableOut,tablesIn...
    ,metadataName)

    % Find missing descriptions:
    missingMeta = tableOut.Properties.VariableNames(cellfun(@isempty...
        ,tableOut.Properties.(metadataName)));
    
    % If there are no missing metadata, return the table as is:
    if isempty(missingMeta)
        return
    end
    
    % Premake a map for collecting available metadata:
    newMeta = containers.Map(missingMeta,repmat({''}...
        ,length(missingMeta),1));
    
    % Collect from tables:
    for curTableIndx = length(tablesIn):-1:1
        
        % Get current table,:
        curTable = tablesIn{curTableIndx};
        
        % Bail if this table does not contain this metadata:
        if isempty(curTable.Properties.(metadataName))
            continue
        end
        
        % Otherwise, check which of the variables from which we need
        % metadata are present:         
        [curPresentVars, metaLocsAll] = ismember(missingMeta...
            ,curTable.Properties.VariableNames);
        
        % Isolate the valid metadata locations:
        metaLocs         = metaLocsAll(curPresentVars);
        presentVarNames  = missingMeta(curPresentVars);
        nonEmptyMetadata = ~cellfun(@isempty...
            ,curTable.Properties.(metadataName)(metaLocs));
        
        % Continue if no valid metadata are present:
        if ~any(nonEmptyMetadata)
            continue
        end
        validMeta      = metaLocs(nonEmptyMetadata);
        validVarNames  = presentVarNames(nonEmptyMetadata);
        
        % Extract the relevant metadata, and descriptions of the relevant
        % variables, generate a map, and overwrite the previous map with
        % new values, if present:
        newMeta = [newMeta;...
            containers.Map(validVarNames...
            ,curTable.Properties.(metadataName)(validMeta))]; %#ok<AGROW>
    end
    
    % Write metadata to table:
    for curVarName = newMeta.keys
        tableOut.Properties.(metadataName)...
            (strcmp(curVarName{1},tableOut.Properties.VariableNames)) ...
            = {newMeta(curVarName{1})};
    end
    
end


