function [epochDataOut] = timeSlice(epochInfo,windowDur)
% timeSlice slices a epochData row in windows with duration d in seconds.
%
%     [epochData] = timeSlice(epochDataRow,epochObj,windowDur)
%
%--------------------------------------------------------------------------
%   Part of the PhysioData Toolbox.
%    Elio Sjak-Shie, Faculty of Social Sciences,
%     Leiden University, 2016.
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Changelog:
%
%   Oct-2016 -- Elio:
%        > Wrote first version.
%
%--------------------------------------------------------------------------

%% Slicer:

% Define necessary precision (0.0001 s):
precisionFixer = 10000;

% Check duration:
sliceDur = str2double(windowDur);
if isnan(sliceDur)||~(sliceDur>0)
    error('The window duration could not be resolved')
end

% Get current epochDataRow:
origEpoch = epochInfo.epochDataRow;

% Fix the precision (a must when using ceil/floor):
origDur = round((origEpoch.endTime-origEpoch.startTime)*precisionFixer)...
    /precisionFixer;

% Calculate the starting and ending time of each epoch:
slicesPerEpoch  = ceil(origDur./sliceDur);
epochSliceStarts = origEpoch.startTime...
    +(0:slicesPerEpoch-1)'.*sliceDur;
epochSliceEnds = (origEpoch.startTime...
    +(1:slicesPerEpoch)'.*sliceDur);
epochSliceEnds(end) = origEpoch.endTime(end);

% Generate slice name:
epochSliceName = strcat(char(origEpoch.epochName)...
    ,{'_'},{'slice_'}...
    , strrep(cellstr(num2str((1:slicesPerEpoch)')),' ',''));

% Strech the original epochData table to fit the new slices, and overwrite
% the necessary data:
epochDataOut               = origEpoch(ones(slicesPerEpoch,1),:);
epochDataOut.epochName     = categorical(epochSliceName);
epochDataOut.startTime     = epochSliceStarts;
epochDataOut.endTime       = epochSliceEnds;
epochDataOut.duration      = epochDataOut.endTime-epochDataOut.startTime;
epochDataOut.epochError    = cell(slicesPerEpoch,1);
epochDataOut.epochMessage  = cell(slicesPerEpoch,1);
epochDataOut.sliceIndex    = (1:slicesPerEpoch)';
epochDataOut.totalSliced   = repmat(slicesPerEpoch,slicesPerEpoch,1);
epochDataOut.parentEpoch   = repmat(origEpoch.epochName,slicesPerEpoch,1);
epochDataOut.Properties.RowNames = strcat(epochSliceName,{' ('}...
    ,char(origEpoch.epochSource),{')'});

% Write desc:
epochDataOut.Properties.VariableDescriptions...
    {strcmp(epochDataOut.Properties.VariableNames,'sliceIndex')}...
    = 'Index of the current slice.';
epochDataOut.Properties.VariableDescriptions...
    {strcmp(epochDataOut.Properties.VariableNames,'totalSliced')}...
    = 'The total amount of slices created from the original epoch';
epochDataOut.Properties.VariableDescriptions...
    {strcmp(epochDataOut.Properties.VariableNames,'parentEpoch')}...
    = 'The original name of the epoch from which the slices were generated.';
epochDataOut.Properties.VariableUnits...
    {strcmp(epochDataOut.Properties.VariableNames,'parentEpoch')}...
    = 'name';



end
